from __future__ import annotations

import inspect

from meltano.core.plugin.base import PluginDefinition

from . import PluginRef


class PluginNotFoundError(Exception):
    """Base exception when a plugin could not be found."""

    def __init__(self, plugin_or_name):
        if isinstance(plugin_or_name, PluginRef):
            self.plugin_type = plugin_or_name.type.descriptor
            self.plugin_name = plugin_or_name.name
        else:
            self.plugin_type = "plugin"
            self.plugin_name = plugin_or_name

    def __str__(self):
        return (
            f"{self.plugin_type.capitalize()} '{self.plugin_name}' is not "
            "known to Meltano"
        )


class PluginParentNotFoundError(Exception):
    """Base exception when a plugin's parent could not be found."""

    def __init__(self, plugin, parent_not_found_error):
        """Initialize exception for when plugin's parent could not be found."""
        self.plugin = plugin
        self.parent_not_found_error = parent_not_found_error

    def __str__(self):
        return (
            f"Could not find parent plugin for {self.plugin.type.descriptor} "
            f"'{self.plugin.name}': {self.parent_not_found_error}"
        )


class PluginNotSupportedError(Exception):
    """Base exception when a plugin is not supported for some operation."""

    def __init__(self, plugin: PluginRef):
        """Construct a PluginNotSupportedError instance."""
        self.plugin = plugin

    def __str__(self):
        return (
            "Operation not supported for "
            f"{self.plugin.type.descriptor} '{self.plugin.name}'"
        )


class PluginExecutionError(Exception):
    """Error stemming from the execution of a plugin (sub-process)."""


class PluginLacksCapabilityError(Exception):
    """Base exception when a plugin lacks a requested capability."""


class InvalidPluginDefinitionError(Exception):
    """Base exception when a plugin definition is invalid."""

    def __init__(self, definition=None):
        """Construct a new "InvalidPluginDefinitionError instance.

        Args:
            definition: Plugin definition.
        """
        fullargspec = inspect.getfullargspec(PluginDefinition.__init__)  # noqa: WPS609
        _, _plugin_type, *required_properties = fullargspec.args

        if isinstance(definition, dict):
            missing_properties = ", ".join(
                [p for p in required_properties if p not in definition],
            )
            self.reason = f"missing properties ({missing_properties})"
        else:
            self.reason = "incorrect format"

    def __str__(self):
        return f"Plugin definition is invalid: {self.reason}"



from __future__ import annotations

import mock
import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.schedule_service import BadCronError
from meltano.core.utils import iso8601_datetime


class TestCliSchedule:
    @pytest.mark.order(0)
    @pytest.mark.usefixtures("project", "session", "tap", "target")
    @mock.patch(
        "meltano.core.schedule_service.PluginSettingsService.get",
        autospec=True,
    )
    def test_schedule_add(self, get, cli_runner, schedule_service):
        test_date = "2010-01-01"
        get.return_value = test_date

        # test adding a scheduled elt
        with mock.patch(
            "meltano.cli.schedule.ScheduleService",
            return_value=schedule_service,
        ):
            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "add",
                    "elt-schedule-mock",
                    "--extractor",
                    "tap-mock",
                    "--loader",
                    "target-mock",
                    "--interval",
                    "@yearly",
                    "--transform",
                    "run",
                ],
            )

        assert_cli_runner(res)
        schedule = schedule_service.schedules()[0]

        assert schedule.name == "elt-schedule-mock"
        assert schedule.extractor == "tap-mock"
        assert schedule.loader == "target-mock"
        assert schedule.transform == "run"
        assert schedule.interval == "@yearly"  # not anytime soon ;)
        assert schedule.start_date == iso8601_datetime(test_date)

        # test adding a scheduled job
        with mock.patch(
            "meltano.cli.schedule.ScheduleService",
            return_value=schedule_service,
        ):
            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "add",
                    "job-schedule-mock",
                    "--job",
                    "mock-job",
                    "--interval",
                    "@yearly",
                ],
            )
        assert res.exit_code == 0

        assert_cli_runner(res)
        schedule = schedule_service.schedules()[1]

        assert schedule.name == "job-schedule-mock"
        assert schedule.job == "mock-job"
        assert schedule.interval == "@yearly"  # not anytime soon ;)

        # Test default schedule case where no argument (set, remove, add, etc)
        # is provided
        with mock.patch(
            "meltano.cli.schedule.ScheduleService",
            return_value=schedule_service,
        ):
            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "job-schedule-mock",
                    "--job",
                    "mock-job",
                    "--interval",
                    "@yearly",
                ],
            )
        assert res.exit_code == 0

        assert_cli_runner(res)
        schedule = schedule_service.schedules()[1]

        assert schedule.name == "job-schedule-mock"
        assert schedule.job == "mock-job"
        assert schedule.interval == "@yearly"  # not anytime soon ;)

        # verify that you can't use job and elt flags together
        with mock.patch(
            "meltano.cli.schedule.ScheduleService",
            return_value=schedule_service,
        ):
            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "add",
                    "job-schedule-mock",
                    "--job",
                    "mock-job",
                    "--extractor",
                    "tap-mock",
                    "--loader",
                    "target-mock",
                    "--interval",
                    "@yearly",
                    "--transform",
                    "run",
                ],
            )

        assert res.exit_code == 1

    @pytest.mark.parametrize("exit_code", (0, 1, 143))
    def test_schedule_run(self, exit_code, cli_runner, elt_schedule, job_schedule):
        process_mock = mock.Mock(returncode=exit_code)
        with mock.patch(
            "meltano.cli.schedule.ScheduleService.run",
            return_value=process_mock,
        ) as run_mock:
            res = cli_runner.invoke(
                cli,
                ["schedule", "run", elt_schedule.name, "--transform", "run"],
            )
            assert res.exit_code == exit_code
            run_mock.assert_called_once_with(elt_schedule, "--transform", "run")

        with mock.patch(
            "meltano.cli.schedule.ScheduleService.run",
            return_value=process_mock,
        ) as run_mock:
            res = cli_runner.invoke(
                cli,
                ["schedule", "run", job_schedule.name, "--dry-run"],
            )
            assert res.exit_code == exit_code
            run_mock.assert_called_once_with(job_schedule, "--dry-run")

    def test_schedule_remove(self, cli_runner, job_schedule):
        process_mock = mock.Mock(returncode=0)

        with mock.patch(
            "meltano.cli.schedule.ScheduleService.remove",
            return_value=process_mock,
        ) as remove_mock:
            res = cli_runner.invoke(cli, ["schedule", "remove", job_schedule.name])
            assert res.exit_code == 0
            remove_mock.assert_called_once_with(job_schedule.name)

    def test_schedule_set(
        self,
        cli_runner,
        elt_schedule,
        job_schedule,
        schedule_service,
    ):
        with mock.patch(
            "meltano.cli.schedule.ScheduleService",
            return_value=schedule_service,
        ):
            res = cli_runner.invoke(
                cli,
                ["schedule", "set", job_schedule.name, "--job", "mock-job-renamed"],
            )
            assert res.exit_code == 0
            assert (
                schedule_service.find_schedule(job_schedule.name).job
                == "mock-job-renamed"
            )

            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "set",
                    elt_schedule.name,
                    "--loader",
                    "mock-target-renamed",
                ],
            )
            assert res.exit_code == 0
            assert (
                schedule_service.find_schedule(elt_schedule.name).loader
                == "mock-target-renamed"
            )

            # interval applies to both and should always work
            res = cli_runner.invoke(
                cli,
                ["schedule", "set", job_schedule.name, "--interval", "@hourly"],
            )
            assert res.exit_code == 0
            assert (
                schedule_service.find_schedule(job_schedule.name).cron_interval
                == "0 * * * *"
            )

            # Should raise exception for invalid cron interval
            res = cli_runner.invoke(
                cli,
                ["schedule", "set", elt_schedule.name, "--interval", "@hourl"],
            )
            assert res.exit_code != 0
            assert isinstance(res.exception, BadCronError)

            res = cli_runner.invoke(
                cli,
                ["schedule", "set", elt_schedule.name, "--interval", "@hourly"],
            )
            assert res.exit_code == 0
            assert (
                schedule_service.find_schedule(elt_schedule.name).cron_interval
                == "0 * * * *"
            )

            # verify that job flags cant be set on elt tasks and vice versa
            res = cli_runner.invoke(
                cli,
                [
                    "schedule",
                    "set",
                    job_schedule.name,
                    "--loader",
                    "mock-target-renamed",
                ],
            )
            assert res.exit_code == 1
            assert (
                schedule_service.find_schedule(job_schedule.name).loader
                != "mock-target-renamed"
            )

            res = cli_runner.invoke(
                cli,
                ["schedule", "set", elt_schedule.name, "--job", "mock-job-renamed"],
            )
            assert res.exit_code == 1
            assert (
                schedule_service.find_schedule(elt_schedule.name).job
                != "mock-job-renamed"
            )



from __future__ import annotations

import shutil
import typing as t

import pytest
from azure.identity import DefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from azure.storage.blob._shared.authentication import (  # noqa: WPS436
    SharedKeyCredentialPolicy,
)

from meltano.core.error import MeltanoError
from meltano.core.state_store import (
    AZStorageStateStoreManager,
    DBStateStoreManager,
    GCSStateStoreManager,
    LocalFilesystemStateStoreManager,
    S3StateStoreManager,
    StateBackend,
    state_store_manager_from_project_settings,
)

if t.TYPE_CHECKING:
    from pathlib import Path

    from meltano.core.project import Project


class TestSystemDBStateBackend:
    def test_manager_from_settings(self, project: Project):
        project.settings.set(["state_backend", "uri"], StateBackend.SYSTEMDB)
        project.settings.set(["state_backend", "lock_timeout_seconds"], 10)
        db_state_store = state_store_manager_from_project_settings(project.settings)
        assert isinstance(db_state_store, DBStateStoreManager)


class TestLocalFilesystemStateBackend:
    @pytest.fixture()
    def state_path(self, tmp_path: Path):
        path = tmp_path / ".meltano" / "state"
        try:
            yield str(path)
        finally:
            shutil.rmtree(path)

    def test_manager_from_settings(self, project: Project, state_path: str):
        project.settings.set(["state_backend", "uri"], f"file://{state_path}")
        file_state_store = state_store_manager_from_project_settings(project.settings)
        assert isinstance(file_state_store, LocalFilesystemStateStoreManager)
        assert file_state_store.state_dir == state_path


class TestAzureStateBackend:
    def test_manager_from_settings(self, project: Project):
        # Azure
        project.settings.set(
            ["state_backend", "uri"],
            "azure://some_container/some/path",
        )
        project.settings.set(
            ["state_backend", "azure", "connection_string"],
            "SOME_CONNECTION_STRING",
        )
        project.settings.set(
            ["state_backend", "azure", "storage_account_url"],
            "SOME_STORAGE_ACCOUNT_URL",
        )
        az_state_store = state_store_manager_from_project_settings(project.settings)
        assert isinstance(az_state_store, AZStorageStateStoreManager)
        assert az_state_store.container_name == "some_container"
        assert az_state_store.prefix == "/some/path"
        assert az_state_store.connection_string == "SOME_CONNECTION_STRING"
        assert az_state_store.storage_account_url == "SOME_STORAGE_ACCOUNT_URL"

        # Azure, missing container name
        project.settings.set(["state_backend", "uri"], "azure://")
        with pytest.raises(MeltanoError):
            state_store_manager_from_project_settings(project.settings)

        # Azure, missing connection string
        project.settings.set(
            ["state_backend", "uri"],
            "azure://some_container/some/path",
        )
        project.settings.unset(["state_backend", "azure", "connection_string"])
        az_state_store = state_store_manager_from_project_settings(project.settings)
        # Should create client using default creds
        assert isinstance(az_state_store, AZStorageStateStoreManager)
        assert isinstance(az_state_store.client, BlobServiceClient)
        assert isinstance(az_state_store.client.credential, DefaultAzureCredential)

        # Azure, missing storage account url
        project.settings.unset(["state_backend", "azure", "storage_account_url"])
        project.settings.set(
            ["state_backend", "azure", "connection_string"],
            (
                "DefaultEndpointsProtocol=https;"
                "AccountName=myAccount;"
                "AccountKey=myAccountKey"
            ),
        )
        az_state_store = state_store_manager_from_project_settings(project.settings)

        # Should create client using connection string
        assert isinstance(az_state_store, AZStorageStateStoreManager)
        assert isinstance(az_state_store.client, BlobServiceClient)
        assert isinstance(az_state_store.client.credential, SharedKeyCredentialPolicy)

        # Azure, missing connection string and storage account url
        project.settings.unset(["state_backend", "azure", "connection_string"])
        project.settings.unset(["state_backend", "azure", "storage_account_url"])
        az_state_store = state_store_manager_from_project_settings(project.settings)

        assert isinstance(az_state_store, AZStorageStateStoreManager)
        # Should raise error
        with pytest.raises(MeltanoError):
            _ = az_state_store.client  # noqa: WPS122


class TestGCSStateBackend:
    def test_manager_from_settings(self, project: Project):
        # GCS
        project.settings.set(["state_backend", "uri"], "gs://some_container/some/path")
        gs_state_store = state_store_manager_from_project_settings(project.settings)
        assert isinstance(gs_state_store, GCSStateStoreManager)
        assert gs_state_store.bucket == "some_container"
        assert gs_state_store.prefix == "/some/path"


class TestS3StateBackend:
    def test_manager_from_settings(self, project: Project):
        # AWS S3 (credentials in URI)
        project.settings.set(
            ["state_backend", "uri"],
            "s3://aws_access_key_id:aws_secret_access_key@some_bucket/some/path",
        )
        s3_state_store = state_store_manager_from_project_settings(project.settings)
        assert isinstance(s3_state_store, S3StateStoreManager)
        assert s3_state_store.bucket == "some_bucket"
        assert s3_state_store.prefix == "/some/path"
        assert s3_state_store.aws_access_key_id == "aws_access_key_id"  # noqa: S105
        assert (
            s3_state_store.aws_secret_access_key == "aws_secret_access_key"  # noqa: S105
        )

        # AWS S3 (credentials provided directly)
        project.settings.set(
            ["state_backend", "uri"],
            "s3://some_bucket/some/path",
        )
        project.settings.set(
            ["state_backend", "s3", "aws_access_key_id"],
            "a_different_id",
        )
        project.settings.set(
            ["state_backend", "s3", "aws_secret_access_key"],
            "a_different_key",
        )
        s3_state_store_direct_creds = state_store_manager_from_project_settings(
            project.settings,
        )
        assert isinstance(s3_state_store_direct_creds, S3StateStoreManager)
        assert s3_state_store_direct_creds.aws_access_key_id == "a_different_id"
        assert (
            s3_state_store_direct_creds.aws_secret_access_key == "a_different_key"  # noqa: S105
        )



"""Plugin glue code for Airflow."""

from __future__ import annotations

import configparser
import os
import subprocess
from contextlib import suppress

import structlog
from packaging.version import Version

from meltano.core.behavior.hookable import hook
from meltano.core.error import AsyncSubprocessError
from meltano.core.plugin import BasePlugin, PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.utils import nest

logger = structlog.stdlib.get_logger(__name__)


class AirflowInvoker(PluginInvoker):
    """Invoker that prepares env for Airflow."""

    def env(self):
        """Environment variables for Airflow.

        Returns:
            Dictionary of environment variables.
        """
        env = super().env()

        env["AIRFLOW_HOME"] = str(self.plugin_config_service.run_dir)
        env["AIRFLOW_CONFIG"] = str(self.files["config"])

        return env


class Airflow(BasePlugin):
    """Plugin glue code for Airflow."""

    __plugin_type__ = PluginType.ORCHESTRATORS

    invoker_class = AirflowInvoker

    @property
    def config_files(self):
        """Return the configuration files required by the plugin.

        Returns:
            Dictionary of config file identifiers and filenames
        """
        return {"config": "airflow.cfg"}

    def process_config(self, flat_config):
        """Unflatten the config.

        Args:
            flat_config: the flat config

        Returns:
            unflattened config
        """
        config = {}
        for key, value in flat_config.items():
            nest(config, key, str(value))
        return config

    @staticmethod
    def update_config_file(invoker: AirflowInvoker) -> None:
        """Update airflow.cfg with plugin configuration.

        Args:
            invoker: the active PluginInvoker
        """
        airflow_cfg_path = invoker.files["config"]
        logger.debug(f"Generated default '{str(airflow_cfg_path)}'")  # noqa: G004

        # open the configuration and update it
        # now we let's update the config to use our stubs
        airflow_cfg = configparser.ConfigParser()

        with airflow_cfg_path.open() as airflow_cfg_file_to_read:
            airflow_cfg.read_file(airflow_cfg_file_to_read)
            logger.debug(f"Loaded '{str(airflow_cfg_path)}'")  # noqa: G004

        config = invoker.plugin_config_processed
        for section, section_config in config.items():
            airflow_cfg[section].update(section_config)
            logger.debug(f"\tUpdated section [{section}] with {section_config}")  # noqa: G004

        with airflow_cfg_path.open("w") as airflow_cfg_file_to_write:
            airflow_cfg.write(airflow_cfg_file_to_write)
            logger.debug(f"Saved '{str(airflow_cfg_path)}'")  # noqa: G004

    @hook("before_install")
    async def setup_env(self, *args, **kwargs):  # noqa: ARG002
        """Configure the env to make airflow installable without GPL deps.

        Args:
            args: Arbitrary args
            kwargs: Arbitrary kwargs
        """
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"

    @hook("before_configure")  # noqa: WPS217
    async def before_configure(self, invoker: AirflowInvoker, session):
        """Generate config file and keep metadata database up-to-date.

        Args:
            invoker: the active PluginInvoker
            session: metadata database session

        Raises:
            AsyncSubprocessError: if command failed to run
        """
        # generate the default `airflow.cfg`
        handle = await invoker.invoke_async(
            "--help",
            require_preparation=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                "Command `airflow --help` failed",  # noqa: EM101
                process=handle,
            )

        # Read and update airflow.cfg
        self.update_config_file(invoker)

        # we've changed the configuration here, so we need to call
        # prepare again on the invoker so it re-reads the configuration
        # for the Airflow plugin
        await invoker.prepare(session)

        # make sure we use correct db init
        handle = await invoker.invoke_async(
            "version",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        stdout, stderr = await handle.communicate()

        if handle.returncode:
            raise AsyncSubprocessError(
                "Command `airflow version` failed",  # noqa: EM101
                process=handle,
            )

        version = stdout.decode()
        init_db_cmd = (
            ["initdb"] if Version(version) < Version("2.0.0") else ["db", "init"]
        )

        handle = await invoker.invoke_async(
            *init_db_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        exit_code = await handle.wait()

        if exit_code:
            raise AsyncSubprocessError(
                (
                    "Airflow metadata database could not be initialized: "  # noqa: EM101
                    "`airflow initdb` failed"
                ),
                handle,
            )

        logger.debug("Completed `airflow initdb`")

    @hook("before_cleanup")
    async def before_cleanup(self, invoker: PluginInvoker):
        """Delete the config file.

        Args:
            invoker: the active PluginInvoker
        """
        config_file = invoker.files["config"]
        with suppress(FileNotFoundError):
            config_file.unlink()
            logger.debug(f"Deleted configuration at {config_file}")  # noqa: G004



"""Use a manifest to establish contexts in which particular configurations are set.

WARNING: This module is currently just here to flesh out an idea for how
         manifests can be used. It is not used anywhere, and may be subject to
         drastic change. Do not rely on its current implementation unless the
         engineering team agrees that this approach is appropriate.
"""

from __future__ import annotations

import os
import typing as t
from contextlib import contextmanager, suppress

if t.TYPE_CHECKING:
    from collections.abc import Mapping

    from meltano.core.manifest.manifest import Manifest

_active_manifest: list[Manifest] = []


def _get_active_manifest() -> Manifest:
    try:
        return _active_manifest[-1]
    except IndexError:
        raise Exception("No manifest has been activated") from None  # noqa: EM101


@contextmanager
def _manifest_context(manifest: Manifest) -> t.Iterator[None]:
    _active_manifest.append(manifest)
    try:
        yield
    finally:
        _active_manifest.pop()


@contextmanager  # noqa: WPS210
def _env_context(env: Mapping[str, str]) -> t.Iterator[None]:
    unique_keys = env.keys() - os.environ.keys()
    shared_keys = env.keys() & os.environ.keys()
    prev = {k: v for k, v in os.environ.items() if k in shared_keys}
    os.environ.update(env)
    try:
        yield
    finally:
        for key in unique_keys:
            with suppress(KeyError):
                del os.environ[key]
        for key in shared_keys:
            os.environ[key] = prev[key]


@contextmanager
def manifest_context(manifest: Manifest) -> t.Iterator[None]:
    """Establish a context within which Meltano can run using a given manifest.

    All relevant general (i.e. excluding plugin-specific, schedule-specific,
    etc.) env vars will be set in the process for the duration of this context
    manager.

    Additionally, this context manager establishes an active manifest, which is
    used by all other context managers within this module.

    Args:
        manifest: The manifest for the current project and environment.

    Yields:
        `None`.
    """
    with _manifest_context(manifest), _env_context(manifest.data["env"]):
        yield


@contextmanager
def plugin_context(plugin_name: str) -> t.Iterator[None]:
    """Establish a context within which a plugin can be run.

    All relevant env vars for the specified plugin will be set in the process
    for the duration of this context manager.

    Args:
        plugin_name: The name of the plugin to be run.

    Raises:
        ValueError: No plugin was found matching the given plugin name within
            the active manifest.

    Yields:
        `None`.
    """
    manifest = _get_active_manifest()

    try:
        plugin = next(
            x
            for v in manifest.data["plugins"].values()
            for x in v
            if x["name"] == plugin_name
        )
    except StopIteration:
        raise ValueError(f"Plugin {plugin_name!r} not found in manifest") from None  # noqa: EM102

    with _env_context(plugin["env"]):
        yield


@contextmanager
def schedule_context(schedule_name: str) -> t.Iterator[None]:
    """Establish a context within which a schedule can be run.

    All relevant env vars for the specified schedule will be set in the process
    for the duration of this context manager.

    Args:
        schedule_name: The name of the schedule to be run.

    Raises:
        ValueError: No schedule was found matching the given schedule name
            within the active manifest.

    Yields:
        `None`.
    """
    schedules = _get_active_manifest().data["schedules"]

    try:
        schedule = next(x for x in schedules if x["name"] == schedule_name)
    except StopIteration:
        raise ValueError(f"Schedule {schedule!r} not found in manifest") from None  # noqa: EM102

    with _env_context(schedule["env"]):
        yield


@contextmanager
def job_context(job_name: str) -> t.Iterator[None]:
    """Establish a context within which a job can be run.

    All relevant env vars for the specified job will be set in the process for
    the duration of this context manager.

    Args:
        job_name: The name of the job to be run.

    Raises:
        ValueError: No job was found matching the given job name within the
            active manifest.

    Yields:
        `None`.
    """
    jobs = _get_active_manifest().data["jobs"]

    try:
        job = next(x for x in jobs if x["name"] == job_name)
    except StopIteration:
        raise ValueError(f"Job {job!r} not found in manifest") from None  # noqa: EM102

    with _env_context(job["env"]):
        yield


# TODO: Provide a function which returns `_get_active_manifest().data` with
# env vars injected into each of its string fields, taking only from
# `os.environ`.



"""Various utilities for configuring logging in a meltano project."""

from __future__ import annotations

import asyncio
import logging
import os
import typing as t
from logging import config as logging_config

import structlog
import yaml

from meltano.core.logging.formatters import (
    LEVELED_TIMESTAMPED_PRE_CHAIN,
    TIMESTAMPER,
    rich_exception_formatter_factory,
)
from meltano.core.utils import get_no_color_flag

if t.TYPE_CHECKING:
    from meltano.core.project import Project

LEVELS: dict[str, int] = {  # noqa: WPS407
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
DEFAULT_LEVEL = "info"
FORMAT = (
    "[%(asctime)s] [%(process)d|%(threadName)10s|%(name)s] [%(levelname)s] %(message)s"  # noqa: WPS323, E501
)


def parse_log_level(log_level: str) -> int:
    """Parse a level descriptor into an logging level.

    Args:
        log_level: level descriptor.

    Returns:
        int: actual logging level.
    """
    return LEVELS.get(log_level, LEVELS[DEFAULT_LEVEL])


def read_config(config_file: os.PathLike | None = None) -> dict | None:
    """Read a logging config yaml from disk.

    Args:
        config_file: path to the config file to read.

    Returns:
        dict: parsed yaml config
    """
    if config_file and os.path.exists(config_file):
        with open(config_file) as cf:
            return yaml.safe_load(cf.read())
    else:
        return None


def default_config(log_level: str) -> dict:
    """Generate a default logging config.

    Args:
        log_level: set log levels to provided level.

    Returns:
         A logging config suitable for use with `logging.config.dictConfig`.
    """
    no_color = get_no_color_flag()

    if no_color:
        formatter = rich_exception_formatter_factory(no_color=True)
    else:
        formatter = rich_exception_formatter_factory(color_system="truecolor")

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.dev.ConsoleRenderer(
                    colors=not no_color,
                    exception_formatter=formatter,
                ),
                "foreign_pre_chain": LEVELED_TIMESTAMPED_PRE_CHAIN,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": log_level.upper(),
                "formatter": "colored",
                "stream": "ext://sys.stderr",
            },
        },
        "loggers": {
            "": {
                "handlers": ["console"],
                "level": log_level.upper(),
                "propagate": True,
            },
            "snowplow_tracker.emitters": {
                "level": logging.ERROR,
            },
            "urllib3": {
                "level": logging.INFO,
            },
            "asyncio": {
                "level": logging.INFO,
            },
            # Azure HTTP logs at info level are too noisy; see
            # https://github.com/meltano/meltano/issues/7723
            "azure.core.pipeline.policies.http_logging_policy": {
                "level": logging.WARNING,
            },
        },
    }


def setup_logging(  # noqa: WPS210
    project: Project | None = None,
    log_level: str = DEFAULT_LEVEL,
    log_config: os.PathLike | None = None,
) -> None:
    """Configure logging for a meltano project.

    Args:
        project: the meltano project
        log_level: set log levels to provided level.
        log_config: a logging config suitable for use with `logging.config.dictConfig`.
    """
    logging.basicConfig(force=True)
    log_level = log_level.upper()

    if project:
        log_config = log_config or project.settings.get("cli.log_config")
        log_level = project.settings.get("cli.log_level")

    config = read_config(log_config) or default_config(log_level)
    logging_config.dictConfig(config)
    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            TIMESTAMPER,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def change_console_log_level(log_level: int = logging.DEBUG) -> None:
    """Change the log level for the root logger, but only on the 'console' handler.

    Most useful when you want change the log level on the fly for console
    output, but want to respect other aspects of any potential `logging.yaml`
    sourced configs. Note that if a `logging.yaml` config without a 'console'
    handler is used, this will not override the log level.

    Args:
        log_level: set log levels to provided level.
    """
    root_logger = logging.getLogger()  # noqa: TID251
    root_logger.setLevel(log_level)
    for handler in root_logger.handlers:
        if handler.name == "console":
            handler.setLevel(log_level)


class SubprocessOutputWriter(t.Protocol):
    """A basic interface suitable for use with `capture_subprocess_output`."""

    def writeline(self, line: str):
        """Write the provided line to an output.

        Args:
            line: String to write
        """


async def _write_line_writer(writer: SubprocessOutputWriter, line: bytes):
    # StreamWriters like a subprocess's stdin need special consideration
    if isinstance(writer, asyncio.StreamWriter):
        try:
            writer.write(line)
            await writer.drain()
        except (BrokenPipeError, ConnectionResetError):
            await writer.wait_closed()
            return False
    else:
        writer.writeline(line.decode(errors="replace"))

    return True


async def capture_subprocess_output(
    reader: asyncio.StreamReader | None,
    *line_writers: SubprocessOutputWriter,
) -> None:
    """Capture in real time the output stream of a subprocess that is run async.

    The stream has been set to asyncio.subprocess.PIPE and is provided using
    reader to this function.

    As new lines are captured for reader, they are written to output_stream.
    This async function should be run with await asyncio.wait() while waiting
    for the subprocess to end.

    Args:
        reader: `asyncio.StreamReader` object that is the output stream of the
            subprocess.
        line_writers: A `StreamWriter`, or object has a compatible writelines method.
    """
    while reader and not reader.at_eof():
        line = await reader.readline()
        if not line:
            continue

        for writer in line_writers:
            if not await _write_line_writer(writer, line):
                # If the destination stream is closed, we can stop capturing output.
                return



"""Meltano Hub command."""

from __future__ import annotations

import typing as t

import click

from meltano.cli.params import pass_project
from meltano.cli.utils import CliEnvironmentBehavior, InstrumentedCmd, InstrumentedGroup
from meltano.core.plugin import PluginType

if t.TYPE_CHECKING:
    from meltano.core.project import Project


@click.group(
    cls=InstrumentedGroup,
    short_help="Interact with Meltano Hub.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
def hub():
    """
    Interact with Meltano Hub.
    Read more at https://docs.meltano.com/reference/command-line-interface#hub
    """  # noqa: D205


@hub.command(
    cls=InstrumentedCmd,
    short_help="Ping Meltano Hub.",
)
@pass_project()
def ping(project: Project):
    """
    Ping Meltano Hub. This can be useful for checking if a custom Hub URL is reachable.
    Read more at https://docs.meltano.com/reference/command-line-interface#hub
    """  # noqa: D205
    try:
        # We want to ensure that we can actually communicate with the Hub.
        # Requesting a list of plugins is a good way to do that, but we don't
        # want to waste bandwidth, so we request the list of orchestrators,
        # which is currently very small.
        project.hub_service.get_plugins_of_type(PluginType.ORCHESTRATORS)
    except Exception as ex:
        raise click.ClickException(
            f"Failed to connect to the Hub at {project.hub_service.hub_api_url!r}",  # noqa: EM102
        ) from ex
    else:
        click.secho(
            f"Successfully connected to the Hub at {project.hub_service.hub_api_url!r}",
            fg="green",
        )



"""Hookable class and supporting functions, classes, and decorators.

This module contains the Hookable class which allows for implementation of a
classic before/after hook pattern. Allowing you to register functions to be
called before or after given trigger.
"""

from __future__ import annotations

from collections import OrderedDict
from contextlib import asynccontextmanager

import structlog

logger = structlog.stdlib.get_logger(__name__)


class hook:  # noqa: N801
    """This decorator marks a function as a `__hook__`.

    It will be found by the `Hookable` metaclass and registered to be triggered
    accordingly.
    """

    def __init__(self, hook_name, can_fail=False):
        self.name = hook_name
        self.can_fail = can_fail

    def __call__(self, func):
        func.__hook__ = self  # noqa: WPS609
        return func


class Hookable(type):
    """Metaclass that registers hook-decorated functions into `__hooks__`.

    Hooks are registered in declaration order.
    """

    def __new__(cls, name, bases, dct):
        new_type = type.__new__(cls, name, bases, dct)  # noqa: WPS609
        new_type.__hooks__ = {}  # noqa: WPS609

        for hook_name, hook in (  # noqa: WPS442
            (func.__hook__.name, func)  # noqa: WPS609, WPS335
            for func in dct.values()
            if hasattr(func, "__hook__")  # noqa: WPS421
        ):
            new_type.__hooks__[hook_name] = new_type.__hooks__.get(  # noqa: WPS609
                hook_name,
                [],
            )  # noqa: WPS609
            new_type.__hooks__[hook_name].append(hook)  # noqa: WPS609

        return new_type

    def __prepare__(cls, bases, **kwds):
        return OrderedDict()


class HookObject(metaclass=Hookable):
    """Hook base class that handles the triggering of hooks.

    Hooks are triggered in reverse MRO order, which means derived classes hooks
    are called after their base class.
    """

    @asynccontextmanager
    async def trigger_hooks(self, hook_name, *args, **kwargs):
        """Trigger all registered before and after functions for a given hook.

        Yields to the caller in between triggers.

        Args:
            hook_name: The hook who's registered functions that should be triggered
            args: Positional arguments to pass to the hooks being triggered.
            kwargs: Keyword arguments to pass to the hooks being triggered.

        Yields:
            `None`

        Examples:
            async with self.obj.trigger_hooks("cleanup", self):
                doStuff()
        """
        try:
            already_triggering = hook_name in self._triggering_hooks
        except AttributeError:
            self._triggering_hooks = set()
            already_triggering = False

        if not already_triggering:
            self._triggering_hooks.add(hook_name)
            await self.__class__.trigger(self, f"before_{hook_name}", *args, **kwargs)

        yield

        if not already_triggering:
            await self.__class__.trigger(self, f"after_{hook_name}", *args, **kwargs)
            self._triggering_hooks.remove(hook_name)

    @classmethod
    async def trigger(cls, target, hook_name, *args, **kwargs):
        """Trigger a registered hook function."""
        hooks = [
            hook
            for hook_cls in reversed(cls.__mro__)
            if hasattr(hook_cls, "__hooks__")  # noqa: WPS421
            for hook in hook_cls.__hooks__.get(hook_name, [])  # noqa: WPS361, WPS609
        ]

        for hook_func in hooks:
            try:
                await hook_func(target, *args, **kwargs)
            except Exception as err:
                if hook_func.__hook__.can_fail:  # noqa: WPS609
                    logger.debug(str(err), exc_info=True)
                    logger.warning(
                        f"{hook_name} hook '{hook_func.__name__}' has failed: {err}",  # noqa: G004
                    )
                else:
                    raise err



from __future__ import annotations

import asyncio
import platform
import signal
import uuid
from datetime import datetime, timedelta, timezone

import pytest

from meltano.core.job.job import (
    HEARTBEAT_VALID_MINUTES,
    HEARTBEATLESS_JOB_VALID_HOURS,
    Job,
    State,
)


class TestJob:
    def sample_job(self, payload=None):
        return Job(
            job_name="meltano:sample-elt",
            state=State.IDLE,
            payload=payload or {},
        )

    def test_save(self, session):
        subject = self.sample_job().save(session)

        assert subject.id > 0

    def test_load(self, session):
        for key in range(10):
            session.add(self.sample_job({"key": key}))

        subjects = session.query(Job).filter_by(job_name="meltano:sample-elt")

        assert len(subjects.all()) == 10
        session.rollback()

    def test_transit(self, session):
        subject = self.sample_job().save(session)

        transition = subject.transit(State.RUNNING)
        assert transition == (State.IDLE, State.RUNNING)
        subject.started_at = datetime.now(timezone.utc)

        transition = subject.transit(State.SUCCESS)
        assert transition == (State.RUNNING, State.SUCCESS)
        subject.ended_at = datetime.now(timezone.utc)

    @pytest.mark.asyncio()
    async def test_run(self, session):
        subject = self.sample_job().save(session)

        # A successful run will mark the subject as SUCCESS and set the `ended_at`
        async with subject.run(session):
            assert subject.state is State.RUNNING
            assert subject.ended_at is None

            await asyncio.sleep(1)
            original_heartbeat = subject.last_heartbeat_at
            assert original_heartbeat is not None

            # Heartbeat is recorded every second
            await asyncio.sleep(2)
            assert subject.last_heartbeat_at > original_heartbeat

            # Yield to give heartbeat another chance to be updated
            await asyncio.sleep(0)

        assert subject.state is State.SUCCESS
        assert subject.ended_at is not None

        # Allow one additional second of delay:
        assert subject.ended_at - subject.last_heartbeat_at < timedelta(seconds=2)

    @pytest.mark.asyncio()
    async def test_run_failed(self, session):
        # A failed run will mark the subject as FAILED an set the payload['error']
        subject = self.sample_job({"original_state": 1}).save(session)
        exception = Exception("This is a test.")

        with pytest.raises(Exception) as err:  # noqa: PT012, PT011
            async with subject.run(session):
                raise exception

            # raise the same exception
            assert err is exception

        assert subject.state is State.FAIL
        assert subject.ended_at is not None
        assert subject.payload["original_state"] == 1
        assert subject.payload["error"] == "This is a test."

    @pytest.mark.asyncio()
    async def test_run_interrupted(self, session):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/2842",
            )
        subject = self.sample_job({"original_state": 1}).save(session)
        with pytest.raises(KeyboardInterrupt):  # noqa: PT012
            async with subject.run(session):
                signal.raise_signal(signal.SIGINT)

        assert subject.state is State.FAIL
        assert subject.ended_at is not None
        assert subject.payload["original_state"] == 1
        assert subject.payload["error"] == "The process was interrupted"

    @pytest.mark.asyncio()
    async def test_run_terminated(self, session):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/2842",
            )
        subject = self.sample_job({"original_state": 1}).save(session)

        with pytest.raises(SystemExit):  # noqa: PT012
            async with subject.run(session):
                signal.raise_signal(signal.SIGTERM)

        assert subject.state is State.FAIL
        assert subject.ended_at is not None
        assert subject.payload["original_state"] == 1
        assert subject.payload["error"] == "The process was terminated"

    def test_run_id(self, session):
        job = Job()
        run_id = job.run_id
        assert isinstance(run_id, uuid.UUID)

        job.save(session)
        assert job.run_id == run_id

    def test_is_stale(self):
        job = Job()

        # Idle jobs are not stale
        assert not job.is_stale()

        # Jobs that were just started are not stale
        job.start()
        assert not job.is_stale()

        # Jobs started more than 25 hours ago without a heartbeat are stale
        offset = timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS + 1)
        job.started_at = datetime.now(timezone.utc) - offset
        assert job.is_stale()

        # Jobs with a recent heartbeat are not stale
        job._heartbeat()
        assert not job.is_stale()

        # Jobs without a heartbeat for 5 minutes are stale
        offset = timedelta(minutes=HEARTBEAT_VALID_MINUTES + 1)
        job.last_heartbeat_at = datetime.now(timezone.utc) - offset
        assert job.is_stale()

        # Completed jobs are not stale
        job.success()
        assert not job.is_stale()

    def test_fail_stale(self):
        job = Job()

        # Leaves a job that isn't stale alone
        assert not job.fail_stale()
        assert not job.has_error()

        # Fails a stale job without a heartbeat
        job.start()
        offset = timedelta(hours=HEARTBEATLESS_JOB_VALID_HOURS + 1)
        job.started_at = datetime.now(timezone.utc) - offset

        assert job.fail_stale()
        assert job.has_error()
        assert "24 hours" in job.payload["error"]

        # Doesn't fail a job that's already failed
        assert not job.fail_stale()

        # Fails a stale job with a heartbeat
        job = Job()
        job.start()
        offset = timedelta(minutes=HEARTBEAT_VALID_MINUTES + 1)
        job.last_heartbeat_at = datetime.now(timezone.utc) - offset

        assert job.fail_stale()
        assert job.has_error()
        assert "5 minutes" in job.payload["error"]



from __future__ import annotations

import json
import platform
import typing as t

import mock
import pytest
from mock import AsyncMock, Mock, patch

from meltano.cli import cli
from meltano.core.plugin import PluginType
from meltano.core.plugin.singer import SingerTap
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.project_plugins_service import ProjectPluginsService

if t.TYPE_CHECKING:
    from click.testing import CliRunner

    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project


@pytest.fixture(scope="class")
def project_tap_mock(project_add_service):
    return project_add_service.add(PluginType.EXTRACTORS, "tap-mock")


@pytest.mark.usefixtures("project_tap_mock")
class TestCliInvoke:
    @pytest.fixture()
    def mock_invoke(self, utility, plugin_invoker_factory):
        process_mock = Mock()
        process_mock.name = "utility-mock"
        process_mock.wait = AsyncMock(return_value=0)

        with patch(
            "meltano.core.plugin_invoker.invoker_factory",
            return_value=plugin_invoker_factory,
        ), patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            invoke_async = AsyncMock(return_value=process_mock)
            asyncio_mock.create_subprocess_exec = invoke_async
            yield invoke_async

    @pytest.fixture()
    def mock_invoke_containers(self, utility, plugin_invoker_factory):
        with patch(
            "meltano.core.plugin_invoker.invoker_factory",
            return_value=plugin_invoker_factory,
        ), patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), mock.patch(
            "aiodocker.Docker",
            autospec=True,
        ) as invoke_async:
            yield invoke_async

    def test_invoke(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "utility-mock"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()
        args, kwargs = mock_invoke.call_args
        assert args[0].endswith("utility-mock")
        assert isinstance(kwargs, dict)

    def test_invoke_args(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "utility-mock", "--help"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()
        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1] == "--help"

    def test_invoke_command(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(
            cli,
            ["invoke", "utility-mock:cmd"],
            env={"ENV_VAR_ARG": "arg"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()

        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1:] == ("--option", "arg")

    def test_invoke_command_containerized(  # noqa: WPS210
        self,
        project,
        cli_runner,
        mock_invoke_containers,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        async def async_generator(*args, **kwargs):  # noqa: ARG001
            yield "Line 1"
            yield "Line 2"  # noqa: WPS354

        docker = mock_invoke_containers.return_value
        docker_context = mock.AsyncMock()
        docker.__aenter__.return_value = docker_context  # noqa: WPS609

        container = mock.AsyncMock()
        docker_context.containers.run.return_value = container

        container.log = mock.Mock()
        container.log.return_value = async_generator()
        container.show.return_value = {"State": {"ExitCode": 0}}

        res = cli_runner.invoke(
            cli,
            ["invoke", "--containers", "utility-mock:containerized"],
            env={"SOME_ENV": "value"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        docker_context.containers.run.assert_called_once()
        container.log.assert_called_once()
        container.wait.assert_called_once()
        container.show.assert_called_once()
        container.delete.assert_called_once()

        args, kwargs = docker_context.containers.run.call_args
        container_config = args[0]
        assert container_config["Cmd"] is None
        assert container_config["Image"] == "mock-utils/mock"
        assert "SOME_ENV=value" in container_config["Env"]

        port_bindings = container_config["HostConfig"]["PortBindings"]
        assert port_bindings == {
            "5000": [{"HostPort": "5000", "HostIP": "0.0.0.0"}],  # noqa: S104
        }

        exposed_ports = container_config["ExposedPorts"]
        assert exposed_ports == {"5000": {}}

        # Check volume env var expansion
        volume_binds = container_config["HostConfig"]["Binds"]
        assert str(project.root) in volume_binds[0]

        assert kwargs["name"].startswith("meltano-utility-mock--containerized")

        # Check env var expansion in volumes
        volume_bindings = args[0]["HostConfig"]["Binds"]
        assert volume_bindings[0].startswith(str(project.root))

    def test_invoke_command_args(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(
            cli,
            ["invoke", "utility-mock:cmd"],
            env={"ENV_VAR_ARG": "arg"},
        )

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_called_once()

        args = mock_invoke.call_args[0]
        assert args[0].endswith("utility-mock")
        assert args[1:] == ("--option", "arg")

    def test_invoke_exit_code(self, cli_runner, mock_invoke):
        mock_invoke.return_value.wait.return_value = 2

        basic = cli_runner.invoke(cli, ["invoke", "utility-mock"])
        assert basic.exit_code == 2

    def test_invoke_triggers(
        self,
        cli_runner: CliRunner,
        project: Project,
        tap: ProjectPlugin,
    ):
        with patch.object(
            SingerTap,
            "discover_catalog",
        ) as discover_catalog, patch.object(
            SingerTap,
            "apply_catalog_rules",
        ) as apply_catalog_rules, patch.object(
            SingerTap,
            "look_up_state",
        ) as look_up_state:
            # Modes other than sync don't trigger discovery or applying catalog rules
            cli_runner.invoke(cli, ["invoke", tap.name, "--some-tap-option"])
            assert discover_catalog.call_count == 0
            assert apply_catalog_rules.call_count == 0
            assert look_up_state.call_count == 0
            project.refresh()

            # Dumping config doesn't trigger discovery or applying catalog rules
            cli_runner.invoke(cli, ["invoke", "--dump", "config", tap.name])
            assert discover_catalog.call_count == 0
            assert apply_catalog_rules.call_count == 0
            assert look_up_state.call_count == 0
            project.refresh()

            # Sync mode triggers discovery and applying catalog rules
            cli_runner.invoke(cli, ["invoke", tap.name])
            assert discover_catalog.call_count == 1
            assert apply_catalog_rules.call_count == 1
            assert look_up_state.call_count == 1
            project.refresh()

            # Dumping catalog triggers discovery and applying catalog rules
            cli_runner.invoke(cli, ["invoke", "--dump", "catalog", tap.name])
            assert discover_catalog.call_count == 2
            assert apply_catalog_rules.call_count == 2
            assert look_up_state.call_count == 2

    def test_invoke_dump_config(self, cli_runner, tap, plugin_settings_service_factory):
        settings_service = plugin_settings_service_factory(tap)

        with patch.object(SingerTap, "discover_catalog"), patch.object(
            SingerTap,
            "apply_catalog_rules",
        ):
            result = cli_runner.invoke(cli, ["invoke", "--dump", "config", tap.name])

            assert json.loads(result.stdout) == settings_service.as_dict(
                extras=False,
                process=True,
            )

    def test_list_commands(self, cli_runner, mock_invoke):
        res = cli_runner.invoke(cli, ["invoke", "--list-commands", "utility-mock"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_invoke.assert_not_called()
        assert "utility-mock:cmd" in res.output
        assert "description of utility command" in res.output

    def test_invoke_only_install(self, cli_runner, project: Project, utility):
        with patch.object(
            ProjectPluginsService,
            "find_plugin",
            return_value=utility,
        ), patch(
            "meltano.cli.params.install_plugins",
        ) as mock_install, patch(
            "meltano.cli.invoke._invoke",
        ) as mock_invoke:
            res = cli_runner.invoke(cli, ["invoke", "--only-install", "utility-mock"])

        assert res.exit_code == 0, f"exit code: {res.exit_code} - {res.exception}"
        mock_install.assert_called_once_with(
            project,
            [utility],
            reason=PluginInstallReason.INSTALL,
        )
        mock_invoke.assert_not_called()



from __future__ import annotations

import json

import pytest

from meltano.core.plugin.singer.catalog import (  # noqa: WPS235
    CatalogRule,
    ListExecutor,
    ListSelectedExecutor,
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    SelectExecutor,
    SelectionType,
    path_property,
    visit,
)

LEGACY_CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "UniqueEntitiesName",
      "stream": "entities",
      "key_properties": [
        "id"
      ],
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "unsupported": {
            "type": "null"
          },
          "payload": {
            "type": "object",
            "properties": {
              "content": {"type": ["string", "null"]},
              "hash": {"type": "string"}
            },
            "required": ["hash"]
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "id"
          ],
          "metadata": {
              "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "code"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "created_at"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "active"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "unsupported"
          ],
          "metadata": {
            "inclusion": "unsupported"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "content"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "hash"
          ],
          "metadata": {
            "inclusion": "available"
          }
        }
      ],
      "replication_key": "created_at",
      "replication_method": "INCREMENTAL"
    }
  ]
}
"""

CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "UniqueEntitiesName",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "unsupported": {
            "type": "null"
          },
          "payload": {
            "type": "object",
            "properties": {
              "content": {"type": ["string", "null"]},
              "hash": {"type": "string"},
              "timestamp": {"type": "string", "format": "date-time"}
            },
            "required": ["hash"]
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "id"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "code"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "created_at"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "active"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "unsupported"
          ],
          "metadata": {
            "inclusion": "unsupported"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "content"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "hash"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "timestamp"
          ],
          "metadata": {
            "inclusion": "available",
            "selected-by-default": true
          }
        },
        {
          "breadcrumb": [],
          "metadata": {
            "table-key-properties": ["id"],
            "valid-replication-keys": ["created_at"],
            "forced-replication-method": "INCREMENTAL"
          }
        }
      ]
    }
  ]
}
"""

JSON_SCHEMA = """
{
  "streams": [
    {
      "tap_stream_id": "UniqueEntitiesName",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "anyOf": [
              {"type": "string", "format": "date-time"},
              {"type": ["string", "null"]}
            ]
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "payload": {
            "type": "object",
            "properties": {
              "content": {"type": ["string", "null"]},
              "hash": {"type": "string"}
            },
            "required": ["hash"]
          }
        }
      }
    }
  ]
}
"""

# Duplicate CATALOG, but change the stream ID to Unique.Entities.Name
# to test that it is properly escaped
ESCAPED_CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "Unique.Entities.Name",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "unsupported": {
            "type": "null"
          },
          "payload": {
            "type": "object",
            "properties": {
              "content": {"type": ["string", "null"]},
              "hash": {"type": "string"},
              "timestamp": {"type": "string", "format": "date-time"}
            },
            "required": ["hash"]
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "id"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "code"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "created_at"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "active"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "unsupported"
          ],
          "metadata": {
            "inclusion": "unsupported"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "content"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "hash"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "payload",
            "properties",
            "timestamp"
          ],
          "metadata": {
            "inclusion": "available",
            "selected-by-default": true
          }
        },
        {
          "breadcrumb": [],
          "metadata": {
            "table-key-properties": ["id"],
            "valid-replication-keys": ["created_at"],
            "forced-replication-method": "INCREMENTAL"
          }
        }
      ]
    }
  ]
}
"""

ESCAPED_JSON_SCHEMA = """
{
  "streams": [
    {
      "tap_stream_id": "Unique.Entities.Name",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "anyOf": [
              {"type": "string", "format": "date-time"},
              {"type": ["string", "null"]}
            ]
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "payload": {
            "type": "object",
            "properties": {
              "content": {"type": ["string", "null"]},
              "hash": {"type": "string"}
            },
            "required": ["hash"]
          }
        }
      }
    }
  ]
}
"""


EMPTY_STREAM_SCHEMA = """
{
  "streams": [
    {
      "tap_stream_id": "UniqueEntitiesName",
      "stream": "entities",
      "schema": {
        "type": "object"
      }
    }
  ]
}
"""

CATALOG_PROPERTIES = {
    "id",
    "code",
    "name",
    "balance",
    "created_at",
    "active",
    "payload",
    "payload.content",
}


@pytest.fixture()
def select_all_executor():
    return SelectExecutor(["*.*"])


@pytest.mark.parametrize(
    ("path", "prop"),
    (
        ("stream[0].properties.master.properties.details", "master.details"),
        ("stream[2].properties.name", "name"),
        ("stream[10].properties.list[2].properties.name", "list[2].name"),
        ("stream[0].properties.properties.properties.prop_1", "properties.prop_1"),
        (
            "stream[0].properties.I HAVE SPACES.properties.SPECIAL CHAR %",
            "I HAVE SPACES.SPECIAL CHAR %",
        ),
    ),
)
def test_path_property(path, prop):
    assert path_property(path) == prop


class TestCatalogRule:
    def test_match(self):
        rule = CatalogRule("tap_stream_id")

        # Stream ID matches
        assert rule.match("tap_stream_id")

        # Stream ID and breadcrumb match
        assert rule.match("tap_stream_id", [])

        # Breadcrumb doesn't match
        assert not rule.match("tap_stream_id", ["property"])

        # Stream ID doesn't match
        assert not rule.match("tap_stream")

    def test_match_wildcard(self):
        rule = CatalogRule("tap_stream*")

        # Stream ID pattern matches
        assert rule.match("tap_stream_id")
        assert rule.match("tap_stream_foo")
        assert rule.match("tap_stream")

        # Stream ID pattern doesn't match
        assert not rule.match("tap_strea")

    def test_match_multiple(self):
        rule = CatalogRule(["tap_stream_id", "other*"])

        # Stream ID patterns match
        assert rule.match("tap_stream_id")
        assert rule.match("other_stream")
        assert rule.match("other_foo")
        assert rule.match("other")

        # Stream ID patterns don't match
        assert not rule.match("tap_stream")
        assert not rule.match("othe")

    def test_match_negated(self):
        rule = CatalogRule("tap_stream_id", negated=True)

        # Stream ID doesn't match, so the rule does
        assert rule.match("tap_stream")

        # Stream ID doesn't match (good!) and breadcrumb does
        assert rule.match("tap_stream", [])

        # Stream ID matches, so the rule doesn't
        assert not rule.match("tap_stream_id")
        assert not rule.match("tap_stream_id", [])
        assert not rule.match("tap_stream_id", ["property"])

        # Stream ID doesn't match (good!), but breadcrumb doesn't match
        assert not rule.match("tap_stream", ["property"])

    def test_match_negated_wildcard(self):
        rule = CatalogRule("tap_stream*", negated=True)

        # Stream ID pattern doesn't match, so the rule does
        assert rule.match("tap_strea")

        # Stream ID pattern matches, so the rule doesn't
        assert not rule.match("tap_stream_id")
        assert not rule.match("tap_stream_foo")
        assert not rule.match("tap_stream")

    def test_match_negated_multiple(self):
        rule = CatalogRule(["tap_stream_id", "other*"], negated=True)

        # Stream ID pattern doesn't match, so the rule does
        assert rule.match("tap_stream")
        assert rule.match("othe")

        # Stream ID pattern matches, so the rule doesn't
        assert not rule.match("tap_stream_id")
        assert not rule.match("other_stream")
        assert not rule.match("other_foo")
        assert not rule.match("other")

    def test_match_breadcrumb(self):
        rule = CatalogRule("tap_stream_id", ["property"])

        # Stream ID matches and breadcrumb is not considered
        assert rule.match("tap_stream_id")

        # Stream ID and breadcrumb match
        assert rule.match("tap_stream_id", ["property"])

        # Stream ID matches, but breadcrumb doesn't
        assert not rule.match("tap_stream_id", [])
        assert not rule.match("tap_stream_id", ["property", "nested"])

    def test_match_wildcard_breadcrumb(self):
        rule = CatalogRule("tap_stream_id", ["proper*"])

        # Stream ID and breadcrumb pattern match
        assert rule.match("tap_stream_id")
        assert rule.match("tap_stream_id", ["property"])
        assert rule.match("tap_stream_id", ["proper_foo"])
        assert rule.match("tap_stream_id", ["proper"])
        assert rule.match("tap_stream_id", ["property", "bar"])

        # Breadcrumb pattern doesn't match
        assert not rule.match("tap_stream_id", [])
        assert not rule.match("tap_stream_id", ["other"])

    def test_match_negated_breadcrumb(self):
        rule = CatalogRule("tap_stream_id", ["property"], negated=True)

        # Stream ID doesn't match (good!) and breadcrumb is not considered
        assert rule.match("tap_stream")

        # Stream ID doesn't match (good!) and breadcrumb does
        assert rule.match("tap_stream", ["property"])

        # Stream ID doesn't match (good!), but breadcrumb doesn't match
        assert not rule.match("tap_stream", [])
        assert not rule.match("tap_stream", ["property", "nested"])


class TestLegacyCatalogSelectVisitor:
    @pytest.fixture()
    def catalog(self):
        return json.loads(LEGACY_CATALOG)

    @classmethod
    def stream_is_selected(cls, stream):
        return stream.get("selected", False)

    @classmethod
    def metadata_is_selected(cls, metadata):
        inclusion = metadata.get("inclusion")
        if inclusion == "automatic":
            return True

        return metadata.get("selected", False)

    @classmethod
    def assert_catalog_is_selected(cls, catalog):
        streams = {stream["tap_stream_id"]: stream for stream in catalog["streams"]}

        # all streams are selected
        for name, stream in streams.items():
            assert cls.stream_is_selected(stream), f"{name} is not selected."

        metadatas = [
            (stream["tap_stream_id"], metadata)
            for _, stream in streams.items()
            for metadata in stream["metadata"]
        ]

        # all fields are selected
        for stream, metadata in metadatas:
            field_metadata = metadata["metadata"]
            if field_metadata.get("inclusion") == "unsupported":
                assert not cls.metadata_is_selected(
                    field_metadata,
                ), f"{stream}.{metadata['breadcrumb']} is selected"
            else:
                assert cls.metadata_is_selected(
                    field_metadata,
                ), f"{stream}.{metadata['breadcrumb']} is not selected"

    def test_visit(self, catalog, select_all_executor):
        visit(catalog, select_all_executor)

        self.assert_catalog_is_selected(catalog)


class TestCatalogSelectVisitor(TestLegacyCatalogSelectVisitor):
    @pytest.fixture()
    def catalog(self, request):
        return json.loads(globals()[request.param])  # noqa: WPS421

    @classmethod
    def stream_is_selected(cls, stream):
        try:
            stream_metadata = next(
                metadata
                for metadata in stream["metadata"]
                if len(metadata["breadcrumb"]) == 0
            )

            return cls.metadata_is_selected(stream_metadata["metadata"])
        except (KeyError, IndexError):
            return False

    @pytest.mark.parametrize(
        "catalog",
        ("CATALOG", "JSON_SCHEMA"),
        indirect=["catalog"],
    )
    def test_visit(self, catalog, select_all_executor):
        super().test_visit(catalog, select_all_executor)

    @pytest.mark.parametrize(
        "catalog",
        ("CATALOG", "JSON_SCHEMA"),
        indirect=["catalog"],
    )
    def test_select_all(self, catalog, select_all_executor):
        visit(catalog, select_all_executor)
        self.assert_catalog_is_selected(catalog)

        streams = {stream["tap_stream_id"]: stream for stream in catalog["streams"]}
        stream_metadata = len(
            [
                metadata
                for stream in streams.values()
                for metadata in stream["metadata"]
                if len(metadata["breadcrumb"]) == 0
            ],
        )

        assert stream_metadata == 1, "Extraneous stream metadata"

    @pytest.mark.parametrize(
        ("catalog", "attrs"),
        (
            (
                "CATALOG",
                {
                    "id",
                    "code",
                    "name",
                    "created_at",
                    "payload.content",
                    "payload.timestamp",
                },
            ),
            ("JSON_SCHEMA", CATALOG_PROPERTIES),
        ),
        indirect=["catalog"],
    )
    def test_select(self, catalog, attrs):
        selector = SelectExecutor(
            [
                "UniqueEntitiesName.code",
                "UniqueEntitiesName.name",
                "*.payload.content",
            ],
        )
        visit(catalog, selector)

        lister = ListSelectedExecutor()
        visit(catalog, lister)

        assert lister.selected_properties["UniqueEntitiesName"] == attrs

    @pytest.mark.parametrize(
        ("catalog", "attrs"),
        (
            (
                "ESCAPED_CATALOG",
                {
                    "id",
                    "code",
                    "name",
                    "created_at",
                    "payload.content",
                    "payload.timestamp",
                },
            ),
            ("ESCAPED_JSON_SCHEMA", CATALOG_PROPERTIES),
        ),
        indirect=["catalog"],
    )
    def test_select_escaped(self, catalog, attrs):
        selector = SelectExecutor(
            [
                "Unique\\.Entities\\.Name.code",
                "Unique\\.Entities\\.Name.name",
                "*.payload.content",
            ],
        )
        visit(catalog, selector)

        lister = ListSelectedExecutor()
        visit(catalog, lister)

        assert lister.selected_properties["Unique.Entities.Name"] == attrs

    @pytest.mark.parametrize(
        ("catalog", "attrs"),
        (
            (
                "CATALOG",
                {"id", "balance", "created_at", "active", "payload", "payload.content"},
            ),
            ("JSON_SCHEMA", CATALOG_PROPERTIES),
        ),
        indirect=["catalog"],
    )
    def test_select_negated(self, catalog, attrs):
        selector = SelectExecutor(
            [
                "*.*",
                "!UniqueEntitiesName.code",
                "!UniqueEntitiesName.name",
                "!*.*.hash",
                "!*.*.timestamp",
            ],
        )
        visit(catalog, selector)

        lister = ListSelectedExecutor()
        visit(catalog, lister)

        assert lister.selected_properties["UniqueEntitiesName"] == attrs

    @pytest.mark.parametrize(
        ("node", "selection_type"),
        (
            (
                {"breadcrumb": ["properties", "a"], "metadata": {"selected": True}},
                SelectionType.SELECTED,
            ),
            (
                {"breadcrumb": ["properties", "a"], "metadata": {"selected": False}},
                SelectionType.EXCLUDED,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"inclusion": "automatic"},
                },
                SelectionType.AUTOMATIC,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"inclusion": "unsupported"},
                },
                SelectionType.EXCLUDED,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"inclusion": "available"},
                },
                SelectionType.EXCLUDED,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"selected": True, "inclusion": "available"},
                },
                SelectionType.SELECTED,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"selected": False, "selected-by-default": True},
                },
                SelectionType.EXCLUDED,
            ),
            (
                {
                    "breadcrumb": ["properties", "a"],
                    "metadata": {"selected-by-default": True},
                },
                SelectionType.SELECTED,
            ),
            (
                {"breadcrumb": ["properties", "a"]},
                SelectionType.EXCLUDED,
            ),
        ),
        ids=[
            "selected: true",
            "selected: false",
            "inclusion: available",
            "selected: true, inclusion: available",
            "inclusion: automatic",
            "inclusion: unsupported",
            "selected: false, selected-by-default: true",
            "selected-by-default: true",
            "no metadata",
        ],
    )
    def test_node_selection(self, node: dict, selection_type: SelectionType):
        """Test that selection metadata produces the expected selection type member."""
        assert ListSelectedExecutor.node_selection(node) == selection_type


class TestSelectionType:
    def test_selection_type_addition(self):
        st = SelectionType
        assert st.EXCLUDED + st.EXCLUDED == st.EXCLUDED
        assert st.SELECTED + st.EXCLUDED == st.EXCLUDED
        assert st.AUTOMATIC + st.EXCLUDED == st.EXCLUDED
        assert st.SELECTED + st.AUTOMATIC == st.AUTOMATIC
        assert st.SELECTED + st.SELECTED == st.SELECTED

    def test_selection_type_repr(self):
        assert f"{SelectionType.EXCLUDED}" == "excluded"
        assert f"{SelectionType.AUTOMATIC}" == "automatic"
        assert f"{SelectionType.SELECTED}" == "selected"


class TestMetadataExecutor:
    @pytest.fixture()
    def catalog(self, request):
        return json.loads(globals()[request.param])  # noqa: WPS421

    @pytest.mark.parametrize(
        "catalog",
        ("CATALOG", "JSON_SCHEMA"),
        indirect=["catalog"],
    )
    def test_visit(self, catalog):
        executor = MetadataExecutor(
            [
                MetadataRule("UniqueEntitiesName", [], "replication-key", "created_at"),
                MetadataRule(
                    "UniqueEntitiesName",
                    ["properties", "created_at"],
                    "is-replication-key",
                    True,
                ),
                MetadataRule(
                    "UniqueEntitiesName",
                    ["properties", "payload", "properties", "hash"],
                    "custom-metadata",
                    "custom-value",
                ),
            ],
        )
        visit(catalog, executor)

        stream_node = next(
            s for s in catalog["streams"] if s["tap_stream_id"] == "UniqueEntitiesName"
        )
        stream_metadata_node = next(
            m for m in stream_node["metadata"] if len(m["breadcrumb"]) == 0
        )
        created_at_property_metadata_node = next(
            m
            for m in stream_node["metadata"]
            if m["breadcrumb"] == ["properties", "created_at"]
        )
        hash_property_metadata_node = next(
            m
            for m in stream_node["metadata"]
            if m["breadcrumb"] == ["properties", "payload", "properties", "hash"]
        )

        assert stream_node["replication_key"] == "created_at"
        assert stream_metadata_node["metadata"]["replication-key"] == "created_at"
        assert (
            created_at_property_metadata_node["metadata"]["is-replication-key"] is True
        )
        assert (
            hash_property_metadata_node["metadata"]["custom-metadata"] == "custom-value"
        )


class TestSchemaExecutor:
    @pytest.fixture()
    def catalog(self, request):
        return json.loads(globals()[request.param])  # noqa: WPS421

    @pytest.mark.parametrize(
        "catalog",
        ("CATALOG", "JSON_SCHEMA", "EMPTY_STREAM_SCHEMA"),
        indirect=["catalog"],
    )
    def test_visit(self, catalog):
        executor = SchemaExecutor(
            [
                SchemaRule(
                    "UniqueEntitiesName",
                    ["properties", "code"],
                    {"anyOf": [{"type": "string"}, {"type": "null"}]},
                ),
                SchemaRule(
                    "UniqueEntitiesName",
                    ["properties", "*_at"],
                    {"type": "string", "format": "date"},
                ),
                SchemaRule(
                    "UniqueEntitiesName",
                    ["properties", "payload"],
                    {
                        "type": "object",
                        "properties": {
                            "content": {"type": ["string", "null"]},
                            "hash": {"type": "string"},
                        },
                    },
                ),
                SchemaRule(
                    "UniqueEntitiesName",
                    ["properties", "*load", "properties", "hash"],
                    {"type": ["string", "null"]},
                ),
            ],
        )
        visit(catalog, executor)

        stream_node = next(
            s for s in catalog["streams"] if s["tap_stream_id"] == "UniqueEntitiesName"
        )
        properties_node = stream_node["schema"]["properties"]

        assert properties_node["code"] == {
            "anyOf": [{"type": "string"}, {"type": "null"}],
        }

        if "created_at" in properties_node:
            assert properties_node["created_at"] == {  # noqa: WPS52
                "type": "string",
                "format": "date",
            }
        else:
            # If no matching properties were found for a glob-like pattern,
            # no new property is created
            assert "*_at" not in properties_node

        assert properties_node["payload"] == {
            "type": "object",
            "properties": {
                "content": {"type": ["string", "null"]},
                "hash": {"type": ["string", "null"]},
            },
        }


class TestListExecutor:
    @pytest.fixture()
    def catalog(self):
        return json.loads(CATALOG)

    def test_visit(self, catalog):
        executor = ListExecutor()
        visit(catalog, executor)

        assert dict(executor.properties) == {
            "UniqueEntitiesName": {
                "code",
                "name",
                "balance",
                "created_at",
                "id",
                "active",
                "unsupported",
                "payload",
                "payload.content",
                "payload.hash",
                "payload.timestamp",
            },
        }



"""Test the compile CLI command."""

from __future__ import annotations

import json
import shutil
import typing as t
from pathlib import Path
from platform import system

import mock
import pytest

from meltano import __file__ as meltano_init_file
from meltano.cli import cli
from meltano.core.manifest import manifest

if t.TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_structlog import StructuredLogCapture

    from meltano.core.project import Project

schema_path = Path(meltano_init_file).parent / "schemas" / "meltano.schema.json"


def check_indent(json_path: Path, indent: int):
    text = json_path.read_text()
    # If the indent is as-expected, then a trip through `json.loads` and
    # `json.dumps` should produce the same text, usually. This isn't true in
    # general, but it's sufficient for our use in the tests here.
    assert json.dumps(json.loads(text), indent=indent if indent > 0 else None) == text


class TestCompile:
    @pytest.fixture()
    def manifest_dir(self, project: Project) -> Path:
        return project.sys_dir_root / "manifests"

    @pytest.fixture(autouse=True)
    def clear_default_manifest_dir(self, manifest_dir: Path) -> None:
        shutil.rmtree(manifest_dir, ignore_errors=True)

    def test_compile(self, manifest_dir: Path, cli_runner: CliRunner):
        assert cli_runner.invoke(cli, ("compile",)).exit_code == 0
        assert {x.name for x in manifest_dir.iterdir()} == {
            f"meltano-manifest{x}.json" for x in (".dev", ".staging", ".prod", "")
        }

    @pytest.mark.parametrize("environment_name", ("dev", "staging", "prod"))
    def test_compile_specific_environment(
        self,
        manifest_dir: Path,
        cli_runner: CliRunner,
        environment_name: str,
    ):
        result = cli_runner.invoke(cli, ("--environment", environment_name, "compile"))
        assert result.exit_code == 0
        assert {x.name for x in manifest_dir.iterdir()} == {
            f"meltano-manifest.{environment_name}.json",
        }

    def test_compile_no_environment(self, manifest_dir: Path, cli_runner: CliRunner):
        result = cli_runner.invoke(cli, ("--no-environment", "compile"))
        assert result.exit_code == 0
        assert {x.name for x in manifest_dir.iterdir()} == {"meltano-manifest.json"}

    @pytest.mark.parametrize("indent", (-3, -1, 0, 1, 4, 9))
    def test_compile_with_indent(
        self,
        manifest_dir: Path,
        cli_runner: CliRunner,
        indent: int,
    ):
        assert (
            cli_runner.invoke(
                cli,
                ("--environment=dev", "compile", f"--indent={indent}"),
            ).exit_code
            == 0
        )
        check_indent(manifest_dir / "meltano-manifest.dev.json", indent)

    def test_specify_output_dir(
        self,
        manifest_dir: Path,
        cli_runner: CliRunner,
        tmp_path: Path,
    ):
        result = cli_runner.invoke(cli, ("compile", "--directory", tmp_path))
        assert result.exit_code == 0
        assert not manifest_dir.exists()
        assert {x.name for x in tmp_path.iterdir()} == {
            f"meltano-manifest{x}.json" for x in (".dev", ".staging", ".prod", "")
        }

    def test_warn_schema_violation(
        self,
        project: Project,
        cli_runner: CliRunner,
        monkeypatch: pytest.MonkeyPatch,
        tmp_path: Path,
        log: StructuredLogCapture,
    ):
        original_yaml_load = manifest.yaml.load

        def patch(*args, **kwargs):
            project_files = original_yaml_load(*args, **kwargs)
            project_files["invalid_key"] = None
            return project_files

        monkeypatch.setenv("NO_COLOR", "1")

        with mock.patch.object(manifest.yaml, "load", side_effect=patch):
            result = cli_runner.invoke(
                cli,
                ("--environment=dev", "compile", "--lint", f"--directory={tmp_path}"),
            )
        assert result.exit_code == 0
        if system() == "Windows":
            # The log message is slightly different on Windows; this length
            # check is good enough:
            assert len(log.events) == 2
            return
        assert log.events == [
            {
                "event": (
                    f"Failed to validate project files against Meltano "
                    f"manifest schema ({schema_path}):\nSchema validation "
                    "errors were encountered.\n  "
                    f"{project.root / 'meltano.yml'}::$: Additional properties"
                    " are not allowed ('invalid_key' was unexpected)"
                ),
                "level": "warning",
            },
            {
                "event": (
                    "Failed to validate newly compiled manifest against "
                    f"Meltano manifest schema ({schema_path}):\nSchema "
                    "validation errors were encountered.\n  "
                    f"{tmp_path / 'meltano-manifest.dev.json'}::$: Additional "
                    "properties are not allowed ('invalid_key' was unexpected)"
                ),
                "level": "warning",
            },
        ]



"""Compile a Meltano manifest."""

from __future__ import annotations

import json
import re
import subprocess
import typing as t
from collections import defaultdict
from contextlib import suppress
from functools import cached_property, reduce
from operator import getitem
from pathlib import Path
from tempfile import NamedTemporaryFile

import flatten_dict
import structlog
import yaml

from meltano import __file__ as package_root_path
from meltano.core.manifest.jsonschema import meltano_config_env_locations
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_lock_service import PluginLock
from meltano.core.utils import (
    EnvVarMissingBehavior,
    MergeStrategy,
    deep_merge,
    default_deep_merge_strategies,
    expand_env_vars,
    get_no_color_flag,
)

if t.TYPE_CHECKING:
    import sys
    from collections.abc import Iterable

    from meltano.core.plugin.base import PluginType
    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project

    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

# NOTE: We do not use `Project(...).meltano.canonical` for 3 reasons:
# - It will make it difficult to refactor the rest of the Meltano core to be
#   based on the manifest if the manifest itself depends on the rest of Meltano
#   core.
# - The canonical representation has changes to the file that we don't want,
#   e.g. turning strings into datetime objects.
# - The canonical representation uses Ruamel, which is known to have poor
#   performance in some cases. No need to eat that cost here, since we don't
#   care about comment preservation in the manifest.

logger = structlog.getLogger(__name__)

JSON_LOCATION_PATTERN = re.compile(r"\.|(\[\])")
MANIFEST_SCHEMA_PATH = (
    Path(package_root_path).parent / "schemas" / "meltano.schema.json"
)

Trie: TypeAlias = t.Dict[str, "Trie"]
PluginsByType: TypeAlias = t.Mapping[str, t.List[t.Mapping[str, t.Any]]]
PluginsByNameByType: TypeAlias = t.Mapping[str, t.Mapping[str, t.Mapping[str, t.Any]]]


# Ruamel doesn't have this problem where YAML tags like timestamps are
# automatically converted to Python objects that aren't JSON-serializable, but
# it's not suitable for use here because it doesn't load the YAML file as plain
# Python data structures like dictionaries and lists. An alternative to using
# `YamlLimitedSafeLoader` would be to load the YAML files with Ruamel, and then
# copying the data out at depth into regular dictionaries and lists.
class YamlLimitedSafeLoader(type):
    """Meta YAML loader that skips the resolution of the specified YAML tags."""

    def __new__(
        cls,
        name,
        bases,
        namespace,
        do_not_resolve: Iterable[str],
    ) -> YamlLimitedSafeLoader:
        """Generate a new instance of this metaclass.

        Args:
            name: The name of the new class.
            bases: The base classes of the new class. The first base class will
                always be `yaml.SafeLoader`. These will come afterwards.
            namespace: The namespace of the new class.
            do_not_resolve: The YAML tags whose resolution should be skipped
                during parsing. They will be left as raw strings.

        Returns:
            The new class object, which is a subclass of `yaml.SafeLoader`
            which skips the resolution of the specified YAML tags.
        """
        do_not_resolve = set(do_not_resolve)
        implicit_resolvers = {
            key: [(tag, regex) for tag, regex in mappings if tag not in do_not_resolve]
            for key, mappings in yaml.SafeLoader.yaml_implicit_resolvers.items()
        }
        return super().__new__(
            cls,
            name,
            tuple({yaml.SafeLoader, *bases}),
            {**namespace, "yaml_implicit_resolvers": implicit_resolvers},
        )


class YamlNoTimestampSafeLoader(
    yaml.SafeLoader,
    metaclass=YamlLimitedSafeLoader,
    do_not_resolve={"tag:yaml.org,2002:timestamp"},
):
    """A safe YAML loader that leaves timestamps as strings."""


class Manifest:  # noqa: WPS214
    """A complete unambiguous static representation of a Meltano project."""

    def __init__(self, project: Project, path: Path, check_schema: bool) -> None:
        """Initialize the manifest.

        This class does not write a manifest file. It merely generates the data
        that would go into one.

        The contents of the manifest can be access using the `data` attribute.
        This data is generated when it is first accessed. After that, a cached
        result is used.

        Args:
            project: The Meltano project which this manifest describes.
            path: The path the Manifest will be saved to - only used for
                jsonschema validation failure error messages.
            check_schema: Whether the project files and generated manifest
                files should be validated against the Meltano schema.
        """
        self.project = project
        self._project_root_str = str(self.project.root.resolve())
        self._meltano_file = self.project.meltanofile.read_text()
        self.path = path
        self.check_schema = check_schema
        with open(MANIFEST_SCHEMA_PATH) as manifest_schema_file:
            manifest_schema = json.load(manifest_schema_file)
        self._env_locations = meltano_config_env_locations(manifest_schema)

    @staticmethod
    def _validate_against_manifest_schema(
        instance_name: str,
        instance_path: Path,
        instance_data: dict[str, t.Any],
    ) -> None:
        with NamedTemporaryFile(suffix=".json") as schema_instance_file:
            schema_instance_file.write(json.dumps(instance_data).encode())
            schema_instance_file.flush()
            proc = subprocess.run(
                (
                    "check-jsonschema",
                    "--color",
                    "never" if get_no_color_flag() else "always",
                    "--schemafile",
                    str(MANIFEST_SCHEMA_PATH),
                    schema_instance_file.name,
                ),
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if proc.returncode:
                jsonschema_checker_message = proc.stdout.strip().replace(
                    schema_instance_file.name,
                    str(instance_path),
                )
                logger.warning(
                    f"Failed to validate {instance_name} against Meltano manifest "  # noqa: G004
                    f"schema ({MANIFEST_SCHEMA_PATH}):\n{jsonschema_checker_message}",
                )

    @cached_property
    def _project_files(self) -> dict[str, t.Any]:
        project_files = flatten_dict.unflatten(
            deep_merge(
                yaml.load(
                    self._meltano_file,
                    t.cast(  # noqa: S506
                        t.Type[yaml.SafeLoader],
                        YamlNoTimestampSafeLoader,
                    ),
                ),
                *(
                    yaml.load(
                        x.read_text(),
                        t.cast(  # noqa: S506
                            t.Type[yaml.SafeLoader],
                            YamlNoTimestampSafeLoader,
                        ),
                    )
                    for x in self.project.project_files.include_paths
                ),
            ),
            "dot",
        )
        if self.check_schema:
            self._validate_against_manifest_schema(
                "project files",
                self.project.meltanofile,
                project_files,
            )
        return project_files

    def _merge_plugin_lockfiles(
        self,
        plugins: dict[PluginType, list[ProjectPlugin]],
        manifest: dict[str, t.Any],
    ) -> None:
        locked_plugins = t.cast(
            t.Dict[str, t.List[t.Mapping[str, t.Any]]],
            {
                plugin_type.value: [
                    PluginLock(self.project, plugin).load(create=True, loader=json.load)
                    for plugin in plugins
                ]
                for (plugin_type, plugins) in plugins.items()
            },
        )

        # Merge the locked plugins with the root-level plugins from `meltano.yml`:
        manifest["plugins"] = deep_merge(  # noqa: WPS204
            _plugins_by_name_by_type(manifest["plugins"]),
            _plugins_by_name_by_type(locked_plugins),
        )

    def sanitize_env_vars(self, env: t.Mapping[str, str]) -> dict[str, str]:
        """Sanitize environment variables.

        Sanitization is performed by:
        - Replacing the project root as an absolute path in values with
            '${MELTANO_PROJECT_ROOT}'

        Args:
            env: A mapping of environment variables.

        Returns:
            The given environment variables with some removed or edited as
            necessary.
        """
        return {
            k: v.replace(self._project_root_str, "${MELTANO_PROJECT_ROOT}")
            for k, v in env.items()
        }

    def env_aware_merge_mappings(
        self,
        data: t.MutableMapping[str, t.Any],
        key: str,
        value: t.Any,
        _: tuple[t.Any, ...] | None = None,
    ) -> None:
        """Merge behavior for `deep_merge` that expands env vars when the key is "env".

        Args:
            data: The dictionary being merged into.
            key: The key for the dictionary being merged into.
            value: The value being merged into the dictionary.
            _: Merge strategies - unused argument.

        Returns:
            `NotImplemented` if the key is not "env"; `None` otherwise.
        """
        if key != "env":
            return NotImplemented  # type: ignore[return-value]
        data[key] = self.sanitize_env_vars(  # noqa: RET503
            {
                **expand_env_vars(
                    data[key],
                    value,
                    if_missing=EnvVarMissingBehavior.ignore,
                ),
                **expand_env_vars(
                    value,
                    data[key],
                    if_missing=EnvVarMissingBehavior.ignore,
                ),
            },
        )

    def _merge_env_vars(
        self,
        plugins: dict[PluginType, list[ProjectPlugin]],
        manifest: dict[str, t.Any],
    ) -> None:
        # Merge env vars derived from project settings:
        self.env_aware_merge_mappings(manifest, "env", self.project.settings.as_env())

        # Ensure the environment-level plugin config is mergable:
        environment = next(iter(manifest["environments"]))
        environment_plugins = environment.setdefault("config", {}).setdefault(
            "plugins",
            {},
        )
        environment["config"]["plugins"] = _plugins_by_name_by_type(environment_plugins)

        merge_strategies = (
            MergeStrategy(dict, self.env_aware_merge_mappings),
            *default_deep_merge_strategies,
        )

        # Env precedence hierarchy:
        # - environment plugin env  # highest
        # - root plugin env
        # - environment env
        # - root env                # Manifest only cares about this level and higher
        # - schedule env
        # - job env
        # - .env file
        # - terminal env            # lowest

        # NOTE: The schedule, job, .env, and terminal env are all injected into
        #       `os.environ`. Meltano should only ever read the schedule and
        #       job env fields within `meltano.yml` or a manifest file in order
        #       to inject it into `os.environ`.

        environment = next(iter(manifest["environments"]))

        # Merge 'environment level plugin env' into 'root level plugin env':
        manifest["plugins"] = deep_merge(
            manifest["plugins"],
            environment.get("config", {}).get("plugins", {}),
            strategies=merge_strategies,
        )

        # Merge env vars derived from plugin settings:
        for plugin_type in plugins:
            for plugin in plugins[plugin_type]:
                self.env_aware_merge_mappings(
                    manifest["plugins"][plugin_type][plugin.name],
                    "env",
                    PluginSettingsService(project=self.project, plugin=plugin).as_env(),
                )

        # Merge 'environment level env' into 'root level env':
        self.env_aware_merge_mappings(manifest, "env", environment["env"])

    @cached_property
    def data(self) -> dict[str, t.Any]:
        """Generate the manifest data.

        The data is generated when this property is first accessed, then cached
        in the instance for fast subsequent accesses. This cached result is
        mutable, because Python lacks a `frozendict` class in its standard
        library. Please do not alter it in-place.

        Returns:
            The manifest data.
        """
        # The manifest begins as the project files, i.e. `meltano.yml` + included files
        manifest = self._project_files

        # Remove all environments other than the one this manifest is for:
        if self.project.environment is None:
            # We use a mock environment here because the rest of the code
            # expects there to be exactly one environment. It is omitted from
            # the final output, and during processing it only affects things if
            # it has content, which this mock does not.
            manifest["environments"] = [{"name": "mock"}]  # noqa: WPS204
        else:
            manifest["environments"] = [
                x
                for x in manifest["environments"]
                if x["name"] == self.project.environment.name
            ]

        apply_scaffold(manifest, self._env_locations)

        plugins = self.project.plugins.plugins_by_type()

        # Remove the 'mappings' category of plugins, since those are created
        # dynamically at run-time, and shouldn't be represented directly in
        # `meltano.yml` or in manifest files. For more details, refer to:
        # https://gitlab.com/meltano/meltano/-/merge_requests/2481#note_832478775
        with suppress(KeyError):
            del plugins["mappings"]

        # NOTE: `self._merge_plugin_lockfiles` restructures the plugins into a
        #       map from plugin types to maps of plugin IDs to their values.
        #       This structure is easier to work with, but is reset to the
        #       proper structure below, before the manifest is returned.
        self._merge_plugin_lockfiles(plugins, manifest)

        self._merge_env_vars(plugins, manifest)

        # Make each plugin type a list again:
        manifest["plugins"] = {
            plugin_type: list(plugins.values())
            for plugin_type, plugins in manifest["plugins"].items()
        }

        # XXX: Meltano shouldn't manipulate annotations, yet here we do
        # because the alternative is to completely lose the environment-level
        # annotations.
        environment = next(iter(manifest["environments"]))
        if "annotations" in environment:
            manifest["annotations"] = deep_merge(
                manifest.get("annotations", {}),
                environment["annotations"],
            )

        # Everything from the selected environment has been merged into the
        # higher levels, so it can be deleted to avoid any ambiguity about
        # which fields within the manifest should be read.
        with suppress(KeyError):
            del manifest["environments"]

        # The include paths have already been resolved, and so are removed to
        # avoid ambiguity. If this information is of interest, it could be
        # re-added into an annotation.
        with suppress(KeyError):
            del manifest["include_paths"]

        if self.check_schema:
            self._validate_against_manifest_schema(
                "newly compiled manifest",
                self.path,
                manifest,
            )

        return manifest


def _plugins_by_name_by_type(
    plugins_by_type: PluginsByType,
) -> PluginsByNameByType:
    return {
        plugin_type: {plugin["name"]: plugin for plugin in plugins}
        for plugin_type, plugins in plugins_by_type.items()
    }


def _locations_trie(paths: Iterable[Iterable[str]]) -> Trie:
    def defaultdict_defaultdict():
        return defaultdict(defaultdict_defaultdict)

    root = defaultdict_defaultdict()
    for path in paths:
        reduce(getitem, path, root)
    return json.loads(json.dumps(root))


def _apply_scaffold(
    trie: Trie,
    manifest_component: dict[str, t.Any] | list[t.Any],
) -> None:
    for key, sub_trie in trie.items():
        if key == "[]":
            if not isinstance(manifest_component, list):
                raise TypeError(
                    "Expected list during manifest scaffolding, "  # noqa: EM102
                    f"got {type(manifest_component)}",
                )
            for element in manifest_component:
                _apply_scaffold(sub_trie, element)
        elif isinstance(manifest_component, dict):
            _apply_scaffold(
                sub_trie,
                manifest_component.setdefault(key, [] if "[]" in sub_trie else {}),
            )
        else:
            raise TypeError(
                "Expected dict during manifest scaffolding, "  # noqa: EM102
                f"got {type(manifest_component)}",
            )


def apply_scaffold(manifest: dict[str, t.Any], locations: Iterable[str]) -> None:
    """Update the provided manifest dict with dicts/lists at the provided locations.

    Args:
        manifest: The manifest dictionary to update in-place.
        locations: The locations (as jq filters) which will be added to the the
            manifest as necessary.
    """
    _apply_scaffold(
        _locations_trie(
            (x for x in JSON_LOCATION_PATTERN.split(location) if x)
            for location in locations
        ),
        manifest,
    )



from __future__ import annotations

import json
import logging
import subprocess
import typing as t
import uuid
from contextlib import contextmanager
from http import server as server_lib
from threading import Thread
from time import sleep

import mock
import pytest
from snowplow_tracker import Emitter, SelfDescribing

from meltano.core.tracking.contexts.cli import CliEvent
from meltano.core.tracking.contexts.environment import EnvironmentContext
from meltano.core.tracking.contexts.exception import ExceptionContext
from meltano.core.tracking.contexts.project import ProjectContext
from meltano.core.tracking.tracker import TelemetrySettings, Tracker
from meltano.core.utils import hash_sha256

if t.TYPE_CHECKING:
    from fixtures.docker import SnowplowMicro
    from meltano.core.project import Project


def load_analytics_json(project: Project) -> dict[str, t.Any]:
    with open(project.meltano_dir() / "analytics.json") as analytics_json_file:
        return json.load(analytics_json_file)


def check_analytics_json(project: Project) -> None:
    analytics_json = load_analytics_json(project)

    for key in ("client_id", "project_id"):
        value = analytics_json[key]
        assert isinstance(value, str)
        uuid.UUID(value)

    assert isinstance(analytics_json["send_anonymous_usage_stats"], bool)


@contextmanager
def delete_analytics_json(project: Project) -> None:
    (project.meltano_dir() / "analytics.json").unlink(missing_ok=True)
    try:
        yield
    finally:
        Tracker(project)


class TestTracker:
    @pytest.fixture(autouse=True)
    def clear_telemetry_settings(self, project, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", raising=False)
        config = project.settings.meltano_yml_config
        config.pop("send_anonymous_usage_stats", None)
        project.settings.update_meltano_yml_config(config)
        original_config_override = (
            type(project.settings).config_override.copy(),
            project.settings.config_override.copy(),
        )
        try:
            type(project.settings).config_override.clear()
            project.settings.config_override.clear()
            project.refresh()
            yield
        finally:
            (  # noqa: WPS414
                type(project.settings).config_override,
                project.settings.config_override,
            ) = original_config_override

    def test_telemetry_state_change_check(self, project: Project):
        with mock.patch.object(
            Tracker,
            "save_telemetry_settings",
        ) as mocked, delete_analytics_json(project):
            Tracker(project)
            assert mocked.call_count == 1

    def test_update_analytics_json(self, project: Project):
        tracker = Tracker(project)
        inital_send_anonymous_usage_stats = tracker.send_anonymous_usage_stats

        # Check the inital value of `send_anonymous_usage_stats`
        analytics_json_pre = load_analytics_json(project)
        assert (
            inital_send_anonymous_usage_stats
            == analytics_json_pre["send_anonymous_usage_stats"]
        )

        # Flip the value of `send_anonymous_usage_stats`
        tracker.send_anonymous_usage_stats = not inital_send_anonymous_usage_stats
        tracker.telemetry_state_change_check(
            TelemetrySettings(
                analytics_json_pre["client_id"],
                analytics_json_pre["project_id"],
                analytics_json_pre["send_anonymous_usage_stats"],
            ),
        )

        # Ensure `send_anonymous_usage_stats` has been flipped on disk
        analytics_json_post = load_analytics_json(project)
        assert (
            inital_send_anonymous_usage_stats
            != analytics_json_post["send_anonymous_usage_stats"]
        )

    def test_restore_project_id_from_analytics_json(self, project: Project):
        Tracker(project)  # Ensure `analytics.json` exists and is valid

        original_project_id = project.settings.get("project_id")

        # Delete the project ID from `meltano.yml`,
        # but leave it unchanged in `analytics.json`
        config = project.settings.meltano_yml_config
        del config["project_id"]
        project.settings.update_meltano_yml_config(config)
        # The project refreshes itself once the config has been changed, which
        # creates a new `ProjectSettingsService`, restoring the `project_id`.
        restored_project_id = project.settings.get("project_id")

        # Depending on what tests were run before this one, the project ID
        # might have been randomly generated, or taken from `analytics.json`,
        # so we accept the restored one if it is equal to the original, or if
        # it is equal after the same transformation that gets applied to the
        # project ID when it is originally stored in `analytics.json`.
        assert original_project_id == restored_project_id or (
            str(uuid.UUID(hash_sha256(original_project_id)[::2])) == restored_project_id
        )

    def test_no_project_id_state_change_if_tracking_disabled(self, project: Project):
        method_name = "track_telemetry_state_change_event"

        project.settings.set("send_anonymous_usage_stats", True)
        project.settings.set("project_id", str(uuid.uuid4()))
        Tracker(project).save_telemetry_settings()

        project.settings.set("send_anonymous_usage_stats", False)
        with mock.patch.object(Tracker, method_name) as mocked:
            Tracker(project).save_telemetry_settings()
            assert mocked.call_count == 1

        project.settings.set("project_id", str(uuid.uuid4()))
        with mock.patch.object(Tracker, method_name) as mocked:
            Tracker(project).save_telemetry_settings()
            assert mocked.call_count == 0

    def test_no_state_change_event_without_analytics_json(self, project: Project):
        method_name = "track_telemetry_state_change_event"

        project.settings.set("send_anonymous_usage_stats", True)
        project.settings.set("project_id", str(uuid.uuid4()))
        Tracker(project).save_telemetry_settings()

        with delete_analytics_json(project):
            project.settings.set("project_id", str(uuid.uuid4()))
            with mock.patch.object(Tracker, method_name) as mocked:
                Tracker(project)
                assert mocked.call_count == 0

    def test_analytics_json_is_created(self, project: Project):
        Tracker(project)
        check_analytics_json(project)
        with delete_analytics_json(project):
            Tracker(project)
            check_analytics_json(project)

    @pytest.mark.parametrize(
        "analytics_json_content",
        (
            f'{{"clientId":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
            f'{{"client_id":"{str(uuid.uuid4())}","projectId":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
            f'{{"client_id":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anon_usage_stats":true}}',  # noqa: E501
            f'["{str(uuid.uuid4())}","{str(uuid.uuid4())}", true]',
            f'client_id":"{str(uuid.uuid4())}","project_id":"{str(uuid.uuid4())}","send_anonymous_usage_stats":true}}',  # noqa: E501
        ),
        ids=(0, 1, 2, 3, 4),
    )
    def test_invalid_analytics_json_is_overwritten(
        self,
        project: Project,
        analytics_json_content: str,
    ):
        with delete_analytics_json(project):
            # Use `delete_analytics_json` to ensure `analytics.json` is restored after
            analytics_json_path = project.meltano_dir() / "analytics.json"
            with open(analytics_json_path, "w") as analytics_json_file:
                analytics_json_file.write(analytics_json_content)

            with pytest.raises((TypeError, KeyError, json.JSONDecodeError)):
                check_analytics_json(project)

            Tracker(project)

            check_analytics_json(project)

    def test_restore_project_id_and_telemetry_state_change(self, project: Project):
        """
        Test that `project_id` is restored from `analytics.json`, and a telemetry state
        change event is fired because `send_anonymous_usage_stats` is negated.
        """  # noqa: D205, D400
        Tracker(project)  # Ensure `analytics.json` exists and is valid

        original_project_id = project.settings.get("project_id")

        # Delete project ID from `meltano.yml`; leave it unchanged in `analytics.json`
        config = project.settings.meltano_yml_config.copy()
        del config["project_id"]
        config["send_anonymous_usage_stats"] = not load_analytics_json(project)[
            "send_anonymous_usage_stats"
        ]
        project.settings.update_meltano_yml_config(config)
        # The project refreshes itself once the config has been changed, which
        # creates a new `ProjectSettingsService`, restoring the `project_id`.
        restored_project_id = project.settings.get("project_id")

        # Depending on what tests were run before this one, the project ID
        # might have been randomly generated, or taken from `analytics.json`,
        # so we accept the restored one if it is equal to the original, or if
        # it is equal after the same transformation that gets applied to the
        # project ID when it is originally stored in `analytics.json`.
        assert original_project_id == restored_project_id or (
            str(uuid.UUID(hash_sha256(original_project_id)[::2])) == restored_project_id
        )

        with mock.patch.object(Tracker, "track_telemetry_state_change_event") as mocked:
            original_config_override = project.settings.config_override.copy()
            try:
                project.settings.config_override.pop("send_anonymous_usage_stats", None)
                Tracker(project)
                assert mocked.call_count == 1
            finally:
                project.settings.config_override = original_config_override

    @pytest.mark.parametrize(
        ("snowplow_endpoints", "send_stats", "expected"),
        (
            (["https://example.com"], True, True),
            (["https://example.com"], False, False),
            ([], True, False),
            ([], False, False),
        ),
    )
    def test_can_track(
        self,
        project: Project,
        snowplow_endpoints: list[str],
        send_stats: bool,
        expected: bool,
    ):
        project.settings.set("snowplow.collector_endpoints", snowplow_endpoints)
        project.settings.set("send_anonymous_usage_stats", send_stats)
        assert Tracker(project).can_track() is expected

    def test_send_anonymous_usage_stats(self, project: Project, monkeypatch):
        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
        assert Tracker(project).send_anonymous_usage_stats is True

        # Ensure the env var takes priority
        project.settings.set("send_anonymous_usage_stats", False)
        assert Tracker(project).send_anonymous_usage_stats is True

        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "False")
        assert Tracker(project).send_anonymous_usage_stats is False

        # Ensure the env var takes priority
        project.settings.set("send_anonymous_usage_stats", True)
        assert Tracker(project).send_anonymous_usage_stats is False

    @pytest.mark.parametrize("setting_value", (False, True))
    def test_send_anonymous_usage_stats_no_env(
        self,
        project: Project,
        setting_value: bool,
    ):
        project.settings.set("send_anonymous_usage_stats", setting_value)
        assert Tracker(project).send_anonymous_usage_stats is setting_value

    def test_default_send_anonymous_usage_stats(self, project: Project):
        assert Tracker(project).send_anonymous_usage_stats

    @pytest.mark.usefixtures("project")
    def test_exit_event_is_fired(self, snowplow: SnowplowMicro):
        subprocess.run(("meltano", "invoke", "alpha-beta-fox"))

        event_summary = snowplow.all()
        assert event_summary["good"] > 0
        assert event_summary["bad"] == 0

        exit_event = next(
            x["event"]
            for x in snowplow.good()
            if x["event"]["event_name"] == "exit_event"
        )
        assert exit_event["unstruct_event"]["data"]["data"]["exit_code"] == 1

    @pytest.mark.parametrize("send_anonymous_usage_stats", (True, False))
    def test_context_with_telemetry_state_change_event(
        self,
        project: Project,
        send_anonymous_usage_stats: bool,
    ):
        tracker = Tracker(project)
        tracker.send_anonymous_usage_stats = send_anonymous_usage_stats

        passed = False

        class MockSnowplowTracker:
            def track(self, event: SelfDescribing):
                # Can't put asserts in here because this method is executed
                # withing a try-except block that catches all exceptions.
                nonlocal passed
                expected_contexts = [EnvironmentContext, ProjectContext]
                if send_anonymous_usage_stats:
                    expected_contexts.append(ExceptionContext)
                passed = len(set(event.context)) == len(expected_contexts) and all(
                    isinstance(ctx, tuple(expected_contexts)) for ctx in event.context
                )

        tracker.snowplow_tracker = MockSnowplowTracker()

        tracker.track_telemetry_state_change_event(
            "project_id",
            uuid.uuid4(),
            uuid.uuid4(),
        )
        assert passed

        tracker.track_telemetry_state_change_event(
            "send_anonymous_usage_stats",
            True,
            False,
        )
        assert passed

        tracker.track_telemetry_state_change_event(
            "send_anonymous_usage_stats",
            False,
            True,
        )
        assert passed

    @pytest.mark.order(0)
    @pytest.mark.parametrize(
        ("sleep_duration", "timeout_should_occur"),
        ((1.0, False), (5.0, True)),
        ids=("no_timeout", "timeout"),
    )
    def test_timeout_if_endpoint_unavailable(
        self,
        project: Project,
        sleep_duration,
        timeout_should_occur,
    ):
        """Test to ensure that the default tracker timeout is respected.

        An HTTP sever is run on another thread, which handles request sent from
        the Snowplow tracker. When `timeout_should_occur` is `False`, we check
        that this sever is properly masquerading as a Snowplow endpoint. When
        `timeout_should_occur` is `True`, we check that when the server takes
        too long to respond, we timeout and continue without raising an error.
        """

        class HTTPRequestHandler(server_lib.SimpleHTTPRequestHandler):
            def do_POST(self) -> None:  # noqa: N802
                sleep(sleep_duration)
                self.send_response(200, "OK")
                self.end_headers()

        server = server_lib.HTTPServer(("localhost", 0), HTTPRequestHandler)
        server_thread = Thread(
            target=server.serve_forever,
            kwargs={"poll_interval": 0.1},
        )

        project.settings.set(
            "snowplow.collector_endpoints",
            f'["http://localhost:{server.server_port}"]',
        )

        tracker = Tracker(project)
        assert len(tracker.snowplow_tracker.emitters) == 1
        tracker.snowplow_tracker.emitters[0].on_failure = mock.MagicMock()

        server_thread.start()
        tracker.track_command_event(CliEvent.started)
        tracker.snowplow_tracker.flush()
        server.shutdown()
        server_thread.join()

        timeout_occurred = (
            tracker.snowplow_tracker.emitters[0].on_failure.call_count  # noqa: WPS219
            == 1
        )
        assert timeout_occurred is timeout_should_occur

    def test_project_context_send_anonymous_usage_stats_source(
        self,
        project: Project,
        monkeypatch,
    ):
        def get_source():
            return ProjectContext(project, uuid.uuid4()).to_json()["data"][
                "send_anonymous_usage_stats_source"
            ]

        assert get_source() == "default"

        project.settings.set(
            "send_anonymous_usage_stats",
            not project.settings.get("send_anonymous_usage_stats"),
        )
        assert get_source() == "meltano_yml"

        monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
        assert get_source() == "env"

    @pytest.mark.order(1)
    def test_get_snowplow_tracker_invalid_endpoint(
        self,
        project: Project,
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
    ):
        endpoints = """
            [
                "notvalid:8080",
                "https://valid.endpoint:8080",
                "file://bad.scheme",
                "https://other.endpoint/path/to/collector"
            ]
        """
        monkeypatch.setenv("MELTANO_SNOWPLOW_COLLECTOR_ENDPOINTS", endpoints)

        with caplog.at_level(logging.WARNING, logger="meltano.core.tracking.tracker"):
            tracker = Tracker(project)

        try:
            assert caplog.records[0].levelname == "WARNING"
            assert caplog.records[0].msg["event"] == "invalid_snowplow_endpoint"
            assert caplog.records[0].msg["endpoint"] == "notvalid:8080"

            assert caplog.records[1].levelname == "WARNING"
            assert caplog.records[1].msg["event"] == "invalid_snowplow_endpoint"
            assert caplog.records[1].msg["endpoint"] == "file://bad.scheme"

            assert len(tracker.snowplow_tracker.emitters) == 2

            emitter = tracker.snowplow_tracker.emitters[0]
            assert isinstance(emitter, Emitter)
            assert emitter.endpoint.startswith(
                "https://valid.endpoint:8080/",
            )

            emitter = tracker.snowplow_tracker.emitters[1]
            assert isinstance(emitter, Emitter)
            assert emitter.endpoint.startswith(
                "https://other.endpoint/path/to/collector/",
            )
        finally:
            # Remove the seemingly valid emitters to prevent a logging error on exit.
            tracker.snowplow_tracker.emitters = []

    def test_client_id_from_env_var(
        self,
        project: Project,
        monkeypatch: pytest.MonkeyPatch,
    ):
        with delete_analytics_json(project):
            monkeypatch.setenv("MELTANO_CLIENT_ID", "invalid-context-uuid")
            with pytest.warns(RuntimeWarning, match="Invalid telemetry client UUID"):
                # Ensure it generated a random UUID as a fallback
                uuid.UUID(str(Tracker(project).client_id))

            ctx_id = uuid.uuid4()
            monkeypatch.setenv("MELTANO_CLIENT_ID", str(ctx_id))
            # Ensure it takes the client ID from the env var
            assert Tracker(project).client_id == ctx_id

            monkeypatch.delenv("MELTANO_CLIENT_ID")
            # Ensure it uses the client ID stored in `analytics.json`
            assert Tracker(project).client_id == ctx_id

            ctx_id_2 = uuid.uuid4()
            monkeypatch.setenv("MELTANO_CLIENT_ID", str(ctx_id_2))
            # Ensure the env var takes priority over `analytics.json`
            assert Tracker(project).client_id == ctx_id_2

            monkeypatch.delenv("MELTANO_CLIENT_ID")
            # Ensure the new client ID overwrites the old one in `analytics.json`
            assert Tracker(project).client_id == ctx_id_2



"""Shared Pytest fixture utilities."""

from __future__ import annotations

import logging
import os
import typing as t
from contextlib import contextmanager

from meltano.core.project import Project
from meltano.core.project_init_service import ProjectInitService

if t.TYPE_CHECKING:
    from pathlib import Path


@contextmanager
def cd(path: Path) -> Path:
    prev_dir = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(prev_dir)


@contextmanager
def tmp_project(name: str, source: Path, compatible_copy_tree) -> Project:
    project_init_service = ProjectInitService(name)
    blank_project = project_init_service.init()
    logging.debug(f"Created new project at {blank_project.root}")  # noqa: G004, TID251
    os.remove(blank_project.meltanofile)
    compatible_copy_tree(source, blank_project.root)
    Project._default = None
    project = Project(blank_project.root)
    with cd(project.root):
        try:
            project.refresh()
            yield project
        finally:
            Project.deactivate()
            logging.debug(f"Cleaned project at {project.root}")  # noqa: G004, TID251



from __future__ import annotations

import json
import typing as t
from signal import SIGTERM

import pytest
from mock import AsyncMock, Mock, patch

from meltano.core.plugin.error import PluginNotSupportedError
from meltano.core.plugin_test_service import (
    ExtractorTestService,
    PluginTestServiceFactory,
)

if t.TYPE_CHECKING:
    from meltano.core.plugin.project_plugin import ProjectPlugin

MOCK_STATE_MESSAGE = json.dumps({"type": "STATE"})
MOCK_RECORD_MESSAGE = json.dumps({"type": "RECORD"})


class TestPluginTestServiceFactory:
    @pytest.fixture(autouse=True)
    @patch("meltano.core.plugin_invoker.PluginInvoker")
    def setup(self, mock_invoker):
        self.mock_invoker = mock_invoker

    def test_extractor_plugin(self, tap: ProjectPlugin):
        self.mock_invoker.plugin = tap

        test_service = PluginTestServiceFactory(self.mock_invoker).get_test_service()
        assert isinstance(test_service, ExtractorTestService)

    def test_loader_plugin(self, target: ProjectPlugin):
        self.mock_invoker.plugin = target

        with pytest.raises(
            PluginNotSupportedError,
            match=(
                f"Operation not supported for {target.type.descriptor} {target.name!r}"
            ),
        ):
            PluginTestServiceFactory(self.mock_invoker).get_test_service()


class TestExtractorTestService:
    @pytest.fixture(autouse=True)
    @patch("meltano.core.plugin_invoker.PluginInvoker")
    def setup(self, mock_invoker):
        self.mock_invoke = Mock()
        self.mock_invoke.name = "utility-mock"
        self.mock_invoke.wait = AsyncMock(return_value=-SIGTERM)
        self.mock_invoke.returncode = -SIGTERM
        self.mock_invoker = mock_invoker
        self.mock_invoker.invoke_async = AsyncMock(return_value=self.mock_invoke)

    @pytest.mark.asyncio()
    async def test_validate_success(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            return_value=b"%b" % MOCK_RECORD_MESSAGE.encode(),
        )

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert is_valid
        assert detail == MOCK_RECORD_MESSAGE

    @pytest.mark.asyncio()
    async def test_validate_success_ignore_non_json(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            side_effect=(b"Not JSON", b"%b" % MOCK_RECORD_MESSAGE.encode()),
        )

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert is_valid
        assert detail == MOCK_RECORD_MESSAGE

    @pytest.mark.asyncio()
    async def test_validate_success_ignore_non_record_msg(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % json.dumps({"key": "value"}).encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert is_valid
        assert detail == MOCK_RECORD_MESSAGE

    @pytest.mark.asyncio()
    async def test_validate_success_stop_after_record_msg(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, False, False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % MOCK_STATE_MESSAGE.encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
                b"%b" % MOCK_RECORD_MESSAGE.encode(),
            ),
        )

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert is_valid
        assert detail == MOCK_RECORD_MESSAGE

        assert self.mock_invoke.stdout.readline.call_count == 2

    @pytest.mark.asyncio()
    async def test_validate_failure_no_record_msg(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            return_value=(b"%b" % MOCK_STATE_MESSAGE.encode()),
        )

        self.mock_invoke.wait = AsyncMock(return_value=0)
        self.mock_invoke.returncode = 0

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert not is_valid
        assert detail is not None
        assert "No RECORD or BATCH message received" in detail

    @pytest.mark.asyncio()
    async def test_validate_failure_subprocess_err(self):
        self.mock_invoke.stderr.at_eof.side_effect = True
        self.mock_invoke.stdout.at_eof.side_effect = (False, False, True)
        self.mock_invoke.stdout.readline = AsyncMock(
            side_effect=(
                b"%b" % MOCK_STATE_MESSAGE.encode(),
                b"A subprocess error occurred",
            ),
        )

        self.mock_invoke.wait = AsyncMock(return_value=1)
        self.mock_invoke.returncode = 1

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert not is_valid
        assert "A subprocess error occurred" in detail

    @pytest.mark.asyncio()
    async def test_validate_failure_plugin_invoke_exception(self):
        mock_exception = Exception("An exception occurred on plugin invocation")
        self.mock_invoker.invoke_async.side_effect = mock_exception

        is_valid, detail = await ExtractorTestService(self.mock_invoker).validate()

        assert not is_valid
        assert str(mock_exception) in detail



"""Base plugin repository."""

from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod

from meltano.core.plugin.factory import base_plugin_factory

if t.TYPE_CHECKING:
    from meltano.core.plugin import BasePlugin, PluginDefinition, PluginType
    from meltano.core.plugin.project_plugin import ProjectPlugin


class PluginRepository(metaclass=ABCMeta):
    """A generic plugin definition repository."""

    @abstractmethod
    def find_definition(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant_name: str | None = None,
    ) -> PluginDefinition:
        """Find a plugin definition.

        Args:
            plugin_type: The type of plugin to find.
            plugin_name: The name of the plugin to find.
            variant_name: The name of the variant to find.
        """

    def find_base_plugin(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        variant: str | None = None,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            plugin_type: The type of plugin to get the base plugin for.
            plugin_name: The name of the plugin to get the base plugin for.
            variant: The variant of the plugin to get the base plugin for.

        Returns:
            The base plugin.
        """
        plugin = self.find_definition(
            plugin_type,
            plugin_name,
        )

        return base_plugin_factory(plugin, variant)

    def get_base_plugin(
        self,
        project_plugin: ProjectPlugin,
        **kwargs,
    ) -> BasePlugin:
        """Get the base plugin for a project plugin.

        Args:
            project_plugin: The project plugin to get the base plugin for.
            kwargs: Additional arguments to pass to the finder.

        Returns:
            The base plugin.
        """
        plugin = project_plugin.custom_definition or self.find_definition(
            project_plugin.type,
            project_plugin.inherit_from or project_plugin.name,
            **kwargs,
        )

        return base_plugin_factory(plugin, project_plugin.variant)



from __future__ import annotations

from datetime import datetime, timedelta, timezone

import pytest
import sqlalchemy as sa

from meltano.core.sqlalchemy import DateTimeUTC


class TestSQLAlchemyModels:
    @pytest.fixture()
    def engine(self) -> sa.Engine:
        return sa.create_engine("sqlite:///:memory:")

    @pytest.fixture()
    def table(self, engine: sa.Engine) -> sa.Table:
        metadata = sa.MetaData()
        test = sa.Table(
            "test",
            metadata,
            sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
            sa.Column("created_at", DateTimeUTC, nullable=True),
        )
        metadata.create_all(engine)
        return test

    @pytest.mark.parametrize(
        ("inserted", "expected"),
        (
            pytest.param(None, None, id="None"),
            pytest.param(
                datetime(2021, 1, 1, 12),  # noqa: DTZ001
                datetime(2021, 1, 1, 12, tzinfo=timezone.utc),
                id="naive",
            ),
            pytest.param(
                datetime(2021, 1, 1, 12, tzinfo=timezone.utc),
                datetime(2021, 1, 1, 12, tzinfo=timezone.utc),
                id="aware",
            ),
            pytest.param(
                datetime(2021, 1, 1, 12, tzinfo=timezone(timedelta(hours=-5))),
                datetime(2021, 1, 1, 17, tzinfo=timezone.utc),
                id="eastern",
            ),
        ),
    )
    def test_datetime_utc_model(
        self,
        engine: sa.Engine,
        table: sa.Table,
        inserted: datetime | None,
        expected: datetime | None,
    ):
        select_stmt = table.select()
        insert_stmt = table.insert().values(created_at=inserted)

        with engine.connect() as conn:
            conn.execute(insert_stmt)

            result = conn.execute(select_stmt).mappings().fetchone()
            assert result is not None
            assert result["created_at"] == expected



from __future__ import annotations


class RunnerError(Exception):
    def __init__(self, message, exitcodes=None):
        super().__init__(message)
        self.exitcodes = {} if exitcodes is None else exitcodes


class Runner:
    def run(self):
        pass



"""Add Subscription

Revision ID: 87d9638f6ac6
Revises: 23ea52e6d784
Create Date: 2020-02-17 16:03:27.765240

"""
import sqlalchemy as sa
from alembic import op

from meltano.migrations import GUID
from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
)

# revision identifiers, used by Alembic.
revision = "87d9638f6ac6"
down_revision = "23ea52e6d784"
branch_labels = None
depends_on = None


def upgrade():
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)

    op.create_table(
        "subscriptions",
        sa.Column("id", GUID, primary_key=True),
        sa.Column("recipient", sa.String(255), nullable=False),
        sa.Column("event_type", sa.String(255), nullable=False),
        sa.Column("source_type", sa.String(255), nullable=True),
        sa.Column("source_id", sa.String(255), nullable=True),
        sa.Column("created_at", datetime_type),
        sa.UniqueConstraint("recipient", "event_type", "source_type", "source_id"),
    )


def downgrade():
    op.drop_table("subscriptions")



from __future__ import annotations

import json
import logging
import platform
import sys
import tempfile
import typing as t

import mock
import pytest
import structlog
from structlog.testing import LogCapture

from meltano.core.logging.output_logger import Out, OutputLogger


def assert_lines(output, *lines):
    for line in lines:
        assert line in output


class TestOutputLogger:
    @pytest.fixture()
    def log(self, tmp_path):
        file = tempfile.NamedTemporaryFile(mode="w+", dir=tmp_path)
        yield file
        file.close()

    @pytest.fixture()
    def subject(self, log):
        return OutputLogger(log.name)

    @pytest.fixture(name="log_output")
    def fixture_log_output(self):
        return LogCapture()

    @pytest.fixture(autouse=True)
    def fixture_configure_structlog(self, log_output):
        original_config = structlog.get_config()
        structlog.configure(
            processors=[log_output],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
        )
        try:
            yield
        finally:
            structlog.configure(**original_config)

    @pytest.fixture(name="redirect_handler")
    def redirect_handler(
        self,
        subject: OutputLogger,
    ) -> t.Generator[logging.Handler, None, None]:
        formatter = structlog.stdlib.ProcessorFormatter(
            # use a json renderer so output is easier to verify
            processor=structlog.processors.JSONRenderer(),
        )
        handler = logging.FileHandler(subject.file)
        handler.setFormatter(formatter)
        yield handler
        handler.close()

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("log")
    async def test_stdio_capture(self, subject, log_output):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        stdout_out = subject.out("stdout")
        stderr_out = subject.out("stderr")

        async with stdout_out.redirect_stdout():
            sys.stdout.write("STD")
            sys.stdout.write("OUT\n")
            print("STDOUT 2")  # noqa: T201, WPS421

        assert_lines(
            log_output.entries,
            {
                "name": "stdout",
                "event": "STDOUT",
                "log_level": "info",
            },
            {
                "name": "stdout",
                "event": "STDOUT 2",
                "log_level": "info",
            },
        )

        async with stderr_out.redirect_stderr():
            sys.stderr.write("STD")
            sys.stderr.write("ERR\n")
            print("STDERR 2", file=sys.stderr)  # noqa: T201, WPS421

        assert_lines(
            log_output.entries,
            {
                "name": "stderr",
                "event": "STDERR",
                "log_level": "info",
            },
            {
                "name": "stderr",
                "event": "STDERR 2",
                "log_level": "info",
            },
        )

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("log")
    async def test_out_writers(self, subject, log_output):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        writer_out = subject.out("writer")
        line_writer_out = subject.out("lwriter")
        basic_out = subject.out("basic")

        async with writer_out.writer() as writer:
            writer.write("WRI")
            writer.write("TER\n")
            writer.write("WRITER 2\n")

        with line_writer_out.line_writer() as line_writer:
            line_writer.write("LINE\n")
            line_writer.write("LINE 2\n")

        basic_out.writeline("LINE\n")
        basic_out.writeline("LINE 2\n")

        assert_lines(
            log_output.entries,
            {
                "name": "writer",
                "event": "WRITER",
                "log_level": "info",
            },
            {
                "name": "writer",
                "event": "WRITER 2",
                "log_level": "info",
            },
            {
                "name": "lwriter",
                "event": "LINE",
                "log_level": "info",
            },
            {
                "name": "lwriter",
                "event": "LINE 2",
                "log_level": "info",
            },
            {"name": "basic", "event": "LINE", "log_level": "info"},
            {
                "name": "basic",
                "event": "LINE 2",
                "log_level": "info",
            },
        )

    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("log")
    async def test_set_custom_logger(self, subject, log_output):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        logger = structlog.getLogger()
        out = subject.out("basic", logger.bind(is_test=True))

        out.writeline("LINE\n")
        assert_lines(
            log_output.entries,
            {
                "name": "basic",
                "event": "LINE",
                "log_level": "info",
                "is_test": True,
            },
        )

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="Test fails if even attempted to be run, xfail can't save us here.",
    )
    @pytest.mark.asyncio()
    @pytest.mark.usefixtures("log", "log_output")
    async def test_logging_redirect(self, subject: OutputLogger, redirect_handler):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        logging_out = subject.out("logging")

        with mock.patch.object(
            Out,
            "redirect_log_handler",
            redirect_handler,
        ), logging_out.redirect_logging():
            logging.info("info")  # noqa: TID251
            logging.warning("warning")  # noqa: TID251
            logging.error("error")  # noqa: TID251

        with open(subject.file) as logf:
            log_file_contents = [json.loads(line) for line in logf.readlines()]

        assert_lines(
            log_file_contents,
            {"event": "info"},
            {"event": "warning"},
            {"event": "error"},
        )

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="Test fails if even attempted to be run, xfail can't save us here.",
    )
    def test_logging_exception(self, log, subject, redirect_handler):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        logging_out = subject.out("logging")

        # it raises logs unhandled exceptions
        exception = Exception("exception")

        with pytest.raises(Exception) as exc, mock.patch.object(  # noqa: PT011
            Out,
            "redirect_log_handler",
            redirect_handler,
        ), logging_out.redirect_logging():
            raise exception

        # make sure it let the exception through
        # All code below here in this test cannot be reached
        assert exc.value is exception

        log_content = json.loads(log.read())

        # make sure the exception is logged
        assert log_content.get("event") == "exception"
        assert log_content.get("exc_info")



from __future__ import annotations

import typing as t
from configparser import ConfigParser

import pytest
from mock import AsyncMock, mock

from meltano.core.plugin import PluginType
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_invoker import asyncio

if t.TYPE_CHECKING:
    from meltano.core.plugin.airflow import AirflowInvoker

AIRFLOW_CONFIG = """

"""


class TestAirflow:
    @pytest.fixture(scope="class")
    def subject(self, project_add_service):
        with mock.patch.object(PluginInstallService, "install_plugin"):
            return project_add_service.add(PluginType.ORCHESTRATORS, "airflow")

    @pytest.mark.asyncio()  # noqa:  WPS210
    async def test_before_configure(  # noqa:  WPS210
        self,
        subject,
        project,
        session,
        plugin_invoker_factory,
    ):
        run_dir = project.run_dir("airflow")

        handle_mock = mock.Mock()
        handle_mock.name = subject.name
        handle_mock.wait = AsyncMock(return_value=0)
        handle_mock.returncode = 0
        handle_mock.communicate = AsyncMock(return_value=(b"2.0.1", None))
        handle_mock.stdout.at_eof.side_effect = (False, True)

        original_exec = asyncio.create_subprocess_exec

        def popen_mock(cmd, *popen_args, **kwargs):
            # first time, it creates the `airflow.cfg`
            if "--help" in popen_args:
                assert "env" in kwargs
                assert kwargs["env"]["AIRFLOW_HOME"] == str(run_dir)

                airflow_cfg = ConfigParser()
                airflow_cfg["core"] = {"dummy": "dummy"}
                airflow_cfg["webserver"] = {"dummy": "dummy"}
                with run_dir.joinpath("airflow.cfg").open("w") as cfg:
                    airflow_cfg.write(cfg)
            # second time, check version
            elif "version" in popen_args:
                return handle_mock
            # third time, it creates the `airflow.db`
            elif {"db", "init"}.issubset(popen_args):
                assert kwargs["env"]["AIRFLOW_HOME"] == str(run_dir)

                project.plugin_dir(subject, "airflow.db").touch()
            else:
                return original_exec(cmd, *popen_args, **kwargs)
            return handle_mock

        with mock.patch.object(
            asyncio,
            "create_subprocess_exec",
            side_effect=popen_mock,
        ) as popen, mock.patch(
            "meltano.core.plugin_invoker.PluginConfigService.configure",
        ) as configure:
            invoker: AirflowInvoker = plugin_invoker_factory(subject)
            # This ends up calling subject.before_configure
            async with invoker.prepared(session):
                commands = [
                    popen_args
                    for _, popen_args, kwargs in popen.mock_calls
                    if popen_args and isinstance(popen_args, tuple)
                ]
                assert commands[0][1] == "--help"
                assert commands[1][1] == "version"
                assert commands[2][1] == "db"
                assert commands[2][2] == "init"
                assert configure.call_count == 2

                assert run_dir.joinpath("airflow.cfg").exists()
                assert project.plugin_dir(subject, "airflow.db").exists()

                airflow_cfg = ConfigParser()
                with run_dir.joinpath("airflow.cfg").open() as cfg:
                    airflow_cfg.read_file(cfg)

                assert airflow_cfg["core"]["dags_folder"]

            assert not run_dir.joinpath("airflow.cfg").exists()

    @pytest.mark.asyncio()
    async def test_before_cleanup(self, subject, plugin_invoker_factory):
        invoker: AirflowInvoker = plugin_invoker_factory(subject)

        assert not invoker.files["config"].exists()

        # No exception should be raised even though the file doesn't exist
        await invoker.cleanup()



from __future__ import annotations

from collections import OrderedDict

import pytest  # noqa: F401

from meltano.core.utils import (
    EnvVarMissingBehavior,
    expand_env_vars,
    flatten,
    makedirs,
    nest,
    pop_at_path,
    remove_suffix,
    set_at_path,
)


def test_nest():
    subject = {}

    one_deep = nest(subject, "a.b")
    one_deep["val"] = 1
    assert one_deep == {"val": 1}

    two_deep = nest(subject, "a.b.c")
    two_deep["val"] = 2
    assert one_deep == {"val": 1, "c": {"val": 2}}

    arr = nest(subject, "a.list", value=[])

    start_value = {"value": 1}
    val = nest(subject, "a.value", value=start_value)

    assert subject["a"]["b"] is one_deep
    assert subject["a"]["b"]["c"] is two_deep
    assert isinstance(arr, list)
    # make sure it is a copy
    assert val == start_value
    assert val is not start_value

    new_b = nest(subject, "a.b", "not_a_dict", force=True)
    assert new_b == "not_a_dict"
    assert subject == {"a": {"b": "not_a_dict", "list": [], "value": {"value": 1}}}

    # make sure existing values aren't cleared when `value=None` and `force=True`
    _ = nest(subject, "a.b", OrderedDict({"d": "d_value"}), force=True)  # noqa: WPS122
    assert subject == {
        "a": {"b": OrderedDict({"d": "d_value"}), "list": [], "value": {"value": 1}},
    }
    similar_b = nest(subject, "a.b", force=True)
    assert similar_b == OrderedDict({"d": "d_value"})
    assert subject == {
        "a": {"b": OrderedDict({"d": "d_value"}), "list": [], "value": {"value": 1}},
    }


def test_pop_at_path():
    subject = {}
    pop_at_path(subject, "a.b.c")
    assert not subject

    subject = {"a": {"b": {"c": "value"}}}
    pop_at_path(subject, "a.b.c")
    assert not subject

    subject = {"a": {"b.c": "value"}}
    pop_at_path(subject, ["a", "b.c"])
    assert not subject

    subject = {"a": {"b": {"c": "value", "d": "value"}, "e": "value"}}
    pop_at_path(subject, "a.b.c")
    assert subject == {"a": {"b": {"d": "value"}, "e": "value"}}

    pop_at_path(subject, "a.b.d")
    assert subject == {"a": {"e": "value"}}

    pop_at_path(subject, "a.e")
    assert not subject


def test_set_at_path():
    subject = {}

    set_at_path(subject, "a.b.c", "value")
    assert subject == {"a": {"b": {"c": "value"}}}

    set_at_path(subject, "a.b.d", "value")
    assert subject == {"a": {"b": {"c": "value", "d": "value"}}}

    set_at_path(subject, "a.b", "value")
    assert subject == {"a": {"b": "value"}}

    set_at_path(subject, "a.b", "newvalue")
    assert subject == {"a": {"b": "newvalue"}}

    set_at_path(subject, "a.b.c", "value")
    assert subject == {"a": {"b": {"c": "value"}}}

    set_at_path(subject, ["a", "d.e"], "value")
    assert subject == {"a": {"b": {"c": "value"}, "d.e": "value"}}


def test_flatten():
    example_config = {"_update": {"orchestrate/dags/meltano.py": False}}
    expected_flat = {"_update.orchestrate/dags/meltano.py": False}
    result = flatten(example_config, "dot")
    assert result == expected_flat


@pytest.mark.parametrize(
    ("input_value", "env", "kwargs", "expected_output"),
    (
        pytest.param(
            "${ENV_VAR}_suffix",
            {"ENV_VAR": "substituted"},
            {},
            "substituted_suffix",
            id="suffix",
        ),
        pytest.param(
            "prefix_${ENV_VAR}",
            {"ENV_VAR": "substituted"},
            {},
            "prefix_substituted",
            id="prefix",
        ),
        pytest.param(
            "prefix_${ENV_VAR}_suffix",
            {"ENV_VAR": "substituted"},
            {},
            "prefix_substituted_suffix",
            id="prefix-and-suffix",
        ),
        pytest.param(
            "${ENV_VAR}",
            {"ENV_VAR": "substituted"},
            {},
            "substituted",
            id="curly-braces",
        ),
        pytest.param(
            "$ENV_VAR",
            {"ENV_VAR": "substituted"},
            {},
            "substituted",
            id="no-curly-braces",
        ),
        pytest.param(
            "$ENV_VAR",
            {},
            {},
            "",
            id="no-match-use-empty-string",
        ),
        pytest.param(
            "$ENV_VAR",
            {},
            {"if_missing": EnvVarMissingBehavior.ignore},
            "$ENV_VAR",
            id="no-match-ignore",
        ),
        pytest.param(
            "${ENV_VAR}",
            {},
            {"if_missing": EnvVarMissingBehavior.ignore},
            "${ENV_VAR}",
            id="no-match-ignore-curly-braces",
        ),
        pytest.param(
            "prefix-${ENV_VAR}-suffix",
            {},
            {"if_missing": EnvVarMissingBehavior.ignore},
            "prefix-${ENV_VAR}-suffix",
            id="no-match-prefix-and-suffix-ignore",
        ),
        pytest.param(
            "MY_DB\\$TableName",
            {},
            {},
            "MY_DB$TableName",
            id="escape",
        ),
    ),
)
def test_expand_env_vars(input_value, env, kwargs, expected_output):
    assert expand_env_vars(input_value, env, **kwargs) == expected_output


def test_expand_env_vars_nested():
    input_dict = {
        "some_key": 12,
        "some_var": "${ENV_VAR_1}",
        "nested": {
            "${THIS_DOES_NOT_EXPAND}": "another_val",
            "another_layer": {"nested_var": "${ENV_VAR_2}"},
        },
        "another_top_level_var": "${ENV_VAR_2}",
    }
    env = {"ENV_VAR_1": "substituted_1", "ENV_VAR_2": "substituted_2"}

    expected_output = {
        "some_key": 12,
        "some_var": "substituted_1",
        "nested": {
            "${THIS_DOES_NOT_EXPAND}": "another_val",
            "another_layer": {"nested_var": "substituted_2"},
        },
        "another_top_level_var": "substituted_2",
    }

    assert expand_env_vars(input_dict, env) == expected_output


@pytest.mark.parametrize(
    ("input_array", "env", "expected_output"),
    (
        pytest.param(
            [
                {"some_key": "${ENV_VAR_1}"},
                {"some_key": "${ENV_VAR_2}"},
            ],
            {"ENV_VAR_1": "substituted_1", "ENV_VAR_2": "substituted_2"},
            [
                {"some_key": "substituted_1"},
                {"some_key": "substituted_2"},
            ],
            id="array-of-flat-dicts",
        ),
        pytest.param(
            [
                {
                    "some_key": [
                        {
                            "some_key": "${ENV_VAR_1}",
                            "another_key": "${ENV_VAR_2}",
                        },
                    ],
                },
            ],
            {"ENV_VAR_1": "substituted_1", "ENV_VAR_2": "substituted_2"},
            [
                {
                    "some_key": [
                        {
                            "some_key": "substituted_1",
                            "another_key": "substituted_2",
                        },
                    ],
                },
            ],
            id="array-of-nested-dicts",
        ),
        pytest.param(
            [
                "${ENV_VAR_1}",
                "${ENV_VAR_2}",
            ],
            {"ENV_VAR_1": "substituted_1", "ENV_VAR_2": "substituted_2"},
            [
                "substituted_1",
                "substituted_2",
            ],
            id="flat-array",
        ),
    ),
)
def test_expand_env_vars_array_nested(input_array, env, expected_output):
    assert expand_env_vars(input_array, env) == expected_output


def test_remove_suffix():
    assert remove_suffix("a_string", "ing") == "a_str"
    assert remove_suffix("a_string", "in") == "a_string"
    assert remove_suffix("a_string", "gni") == "a_string"


def test_makedirs_decorator(tmp_path):
    def root(*paths):
        return tmp_path.joinpath(*paths)

    @makedirs
    def hierarchy(*ranks, make_dirs: bool = True):  # noqa: ARG001
        return root(*ranks)

    @makedirs
    def species(genus_name, species_name, make_dirs: bool = True):  # noqa: ARG001
        return hierarchy(genus_name, species_name, make_dirs=make_dirs)

    cat = species("felis", "catus")
    assert cat.exists()

    wolf = species("canis", "lupus", make_dirs=False)
    assert not wolf.exists()
